"""Test the ratio module."""
from particula.activity.ratio import to_molar_mass_ratio, from_molar_mass_ratio


def test_to_molar_mass_ratio():
    """Test the to_molar_mass_ratio function."""
    # Test case 1: Single molar mass
    molar_mass = 30.0
    expected_ratio = 18.01528 / molar_mass
    assert to_molar_mass_ratio(molar_mass) == expected_ratio

    # Test case 2: List of molar masses
    molar_masses = [20.0, 40.0, 60.0]
    expected_ratios = [18.01528 / mm for mm in molar_masses]
    assert to_molar_mass_ratio(molar_masses) == expected_ratios


def test_from_molar_mass_ratio():
    """Test the from_molar_mass_ratio function."""
    # Test case 1: Single molar mass ratio
    molar_mass_ratio = 0.5
    expected_molar_mass = 18.01528 * molar_mass_ratio
    assert from_molar_mass_ratio(molar_mass_ratio) == expected_molar_mass

    # Test case 2: List of molar mass ratios
    molar_mass_ratios = [0.2, 0.4, 0.6]
    expected_molar_masses = [18.01528 * mm for mm in molar_mass_ratios]
    assert from_molar_mass_ratio(molar_mass_ratios) == expected_molar_masses



""" testing the mean free path calculation
"""

import pytest
import numpy as np
from particula.gas.properties import molecule_mean_free_path


def test_molecule_mean_free_path():
    """Testing the mean free path of a molecule compare with mfp"""

    a_mfp = 6.52805868e-08  # at stp
    b_molecule_mfp = molecule_mean_free_path(
        temperature=298, pressure=101325, molar_mass=0.03
    )
    assert pytest.approx(a_mfp, rel=1e-6) == b_molecule_mfp


def test_dynamic_viscosity_provided():
    """Test when dynamic viscosity is explicitly provided"""
    dynamic_viscosity = 5 * 1.78e-5  # 5x Value for air at stp
    result = molecule_mean_free_path(dynamic_viscosity=dynamic_viscosity)
    assert result > 0


def test_array_input():
    """Test when array inputs are provided for temperature, pressure,
    and molar mass"""
    molar_masses = np.array([0.028, 0.044])
    results = molecule_mean_free_path(molar_mass=molar_masses)
    # All calculated mean free paths should be positive
    assert np.all(results > 0)


@pytest.mark.parametrize("temperature", [None, -1, "a"])
def test_invalid_temperature(temperature):
    """Test when invalid temperature values are provided to the function"""
    with pytest.raises((TypeError, ValueError)):
        molecule_mean_free_path(temperature=temperature)


@pytest.mark.parametrize("pressure", [None, -1, "a"])
def test_invalid_pressure(pressure):
    """Test when invalid pressure values are provided to the function"""
    with pytest.raises((TypeError, ValueError)):
        molecule_mean_free_path(pressure=pressure)


@pytest.mark.parametrize(
    "molar_mass", [None, -1, "a", np.array([0.028, -0.044])]
)
def test_invalid_molar_mass(molar_mass):
    """Test when invalid molar mass values are provided to the function"""
    with pytest.raises((TypeError, ValueError)):
        molecule_mean_free_path(molar_mass=molar_mass)



"""
Fit coefficients for the Binary Activity Coefficient
model for organic-water mixtures.

Gorkowski, K., Preston, T. C., &#38; Zuend, A. (2019).
Relative-humidity-dependent organic aerosol thermodynamics
Via an efficient reduced-complexity model.
Atmospheric Chemistry and Physics
https://doi.org/10.5194/acp-19-13383-2019
"""

from typing import Union, List, NamedTuple
import numpy as np
from numpy.typing import NDArray


class FitValues(NamedTuple):
    """
    Named tuple for the fit values for the activity model.
    """
    a1: List[float]
    a2: List[float]
    s: List[float]


G19_FIT_LOW = FitValues(
    a1=[7.089476e00, -7.711860e00, -3.885941e01, -1.000000e02],
    a2=[-6.226781e-01, -1.000000e02, 3.081244e-09, 6.188812e01],
    s=[-5.988895e00, 6.940689e00],
)
G19_FIT_MID = FitValues(
    a1=[5.872214e00, -4.535007e00, -5.129327e00, -2.809232e01],
    a2=[-9.740486e-01, -1.000000e02, 2.109751e00, -2.367683e01],
    s=[-1.219164e00, 4.742729e00],
)
G19_FIT_HIGH = FitValues(
    a1=[5.921550e00, -2.528295e00, -3.883017e00, -7.898128e00],
    a2=[-1.000000e02, -1.000000e02, 1.353916e00, -1.160145e01],
    s=[-7.868187e-02, 3.650860e00],
)
INTERPOLATE_WATER_FIT = 500
LOWEST_ORGANIC_MOLE_FRACTION = 1e-12


def coefficients_c(
    molar_mass_ratio: Union[float, NDArray[np.float64]],
    oxygen2carbon: Union[float, NDArray[np.float64]],
    fit_values: List[float],
) -> NDArray[np.float64]:
    """
    Coefficients for activity model, see Gorkowski (2019). equation S1 S2.

    Args:
        - molar_mass_ratio : The molar mass ratio of water to organic
          matter.
        - oxygen2carbon : The oxygen to carbon ratio.
        - fit_values : The fit values for the activity model.

    Returns:
        - The coefficients for the activity model.
    """
    # force to array
    molar_mass_ratio = np.asarray(molar_mass_ratio, dtype=np.float64)
    oxygen2carbon = np.asarray(oxygen2carbon, dtype=np.float64)
    fit_values = np.asarray(fit_values, dtype=np.float64)

    return fit_values[0] * np.exp(fit_values[1] * oxygen2carbon) + fit_values[
        2
    ] * np.exp(fit_values[3] * molar_mass_ratio)



"""Builder class for Activity objects with validation and error handling.

Change to MixinMolar classes, after PR integration.
"""

import logging
from typing import Optional, Union
from numpy.typing import NDArray
import numpy as np
from particula.abc_builder import (
    BuilderABC,
)
from particula.builder_mixin import (
    BuilderMolarMassMixin,
    BuilderDensityMixin,
)
from particula.particles.activity_strategies import (
    ActivityIdealMass,
    ActivityIdealMolar,
    ActivityKappaParameter,
    ActivityStrategy,
)

logger = logging.getLogger("particula")


class ActivityIdealMassBuilder(BuilderABC):
    """Builder class for IdealActivityMass objects. No additional parameters.

    Methods:
        build(): Validate and return the IdealActivityMass object.
    """

    def __init__(self):
        required_parameters = None
        BuilderABC.__init__(self, required_parameters)

    def build(self) -> ActivityStrategy:
        """Validate and return the IdealActivityMass object.

        Returns:
            IdealActivityMass: The validated IdealActivityMass object.
        """
        return ActivityIdealMass()


class ActivityIdealMolarBuilder(BuilderABC, BuilderMolarMassMixin):
    """Builder class for IdealActivityMolar objects.

    Methods:
        set_molar_mass(molar_mass, molar_mass_units): Set the molar mass of the
            particle in kg/mol. Default units are 'kg/mol'.
        set_parameters(params): Set the parameters of the IdealActivityMolar
            object from a dictionary including optional units.
        build(): Validate and return the IdealActivityMolar object.
    """

    def __init__(self):
        required_parameters = ["molar_mass"]
        BuilderABC.__init__(self, required_parameters)
        BuilderMolarMassMixin.__init__(self)

    def build(self) -> ActivityStrategy:
        """Validate and return the IdealActivityMolar object.

        Returns:
            IdealActivityMolar: The validated IdealActivityMolar object.
        """
        self.pre_build_check()
        return ActivityIdealMolar(molar_mass=self.molar_mass)  # type: ignore


class ActivityKappaParameterBuilder(
    BuilderABC, BuilderDensityMixin, BuilderMolarMassMixin
):
    """Builder class for KappaParameterActivity objects.

    Methods:
        set_kappa(kappa): Set the kappa parameter for the activity calculation.
        set_density(density,density_units): Set the density of the species in
            kg/m^3. Default units are 'kg/m^3'.
        set_molar_mass(molar_mass,molar_mass_units): Set the molar mass of the
            species in kg/mol. Default units are 'kg/mol'.
        set_water_index(water_index): Set the array index of the species.
        set_parameters(dict): Set the parameters of the KappaParameterActivity
            object from a dictionary including optional units.
        build(): Validate and return the KappaParameterActivity object.
    """

    def __init__(self):
        required_parameters = ["kappa", "density", "molar_mass", "water_index"]
        BuilderABC.__init__(self, required_parameters)
        BuilderDensityMixin.__init__(self)
        BuilderMolarMassMixin.__init__(self)
        self.kappa = None
        self.water_index = None

    def set_kappa(
        self,
        kappa: Union[float, NDArray[np.float64]],
        kappa_units: Optional[str] = None,
    ):
        """Set the kappa parameter for the activity calculation.

        Args:
            kappa: The kappa parameter for the activity calculation.
            kappa_units: Not used. (for interface consistency)
        """
        if np.any(kappa < 0):
            error_message = "Kappa parameter must be a positive value."
            logger.error(error_message)
            raise ValueError(error_message)
        if kappa_units is not None:
            logger.warning("Ignoring units for kappa parameter.")
        self.kappa = kappa
        return self

    def set_water_index(
        self, water_index: int, water_index_units: Optional[str] = None
    ):
        """Set the array index of the species.

        Args:
            water_index: The array index of the species.
            water_index_units: Not used. (for interface consistency)
        """
        if not isinstance(water_index, int):  # type: ignore
            error_message = "Water index must be an integer."
            logger.error(error_message)
            raise TypeError(error_message)
        if water_index_units is not None:
            logger.warning("Ignoring units for water index.")
        self.water_index = water_index
        return self

    def build(self) -> ActivityStrategy:
        """Validate and return the KappaParameterActivity object.

        Returns:
            KappaParameterActivity: The validated KappaParameterActivity
                object.
        """
        self.pre_build_check()
        return ActivityKappaParameter(
            kappa=self.kappa,  # type: ignore
            density=self.density,  # type: ignore
            molar_mass=self.molar_mass,  # type: ignore
            water_index=self.water_index,  # type: ignore
        )



"""Test for collision_radius_module.py module."""

import numpy as np
from particula.particles.properties.collision_radius_module import (
    mulholland_1988,
    rogak_flagan_1992,
    zurita_gotor_2002,
    thajudeen_2012,
    qian_2022_rg,
    qian_2022_rg_df,
    qian_2022_rg_df_k0,
    qian_2022_rg_df_k0_a13,
)


def test_mulholland_1988():
    """Test the mulholland_1988 function."""
    assert mulholland_1988(1.5) == 1.5
    assert mulholland_1988(0.0) == 0.0
    assert mulholland_1988(-1.5) == -1.5
    assert np.isnan(mulholland_1988(np.nan))
    assert np.isinf(mulholland_1988(np.inf))
    test_array = np.array([[1.5, 0.0], [-1.5, 2.0]])
    np.testing.assert_array_equal(mulholland_1988(test_array), test_array)


def test_rogak_flagan_1992():
    """Test the rogak_flagan_1992 function."""
    result = rogak_flagan_1992(1.5, 2.5)
    expected = np.sqrt((2.5 + 2) / 3) * 1.5
    assert np.isclose(result, expected)


def test_zurita_gotor_2002():
    """Test the zurita_gotor_2002 function."""
    result = zurita_gotor_2002(1.5, 1.2)
    expected = 1.037 * (1.2**0.077) * 1.5
    assert np.isclose(result, expected)


def test_thajudeen_2012():
    """Test the thajudeen_2012 function."""
    result = thajudeen_2012(2.5, 100, 1.5, 0.1)
    alpha1 = 0.253 * 2.5**2 - 1.209 * 2.5 + 1.433
    alpha2 = -0.218 * 2.5**2 + 0.964 * 2.5 - 0.180
    phi = 1 / (alpha1 * np.log(100) + alpha2)
    radius_s_i = phi * 1.5
    radius_s_ii = (0.1 * (1.203 - 0.4315 / 2.5) / 2) * (
        4 * radius_s_i / 0.1
    ) ** (0.8806 + 0.3497 / 2.5)
    expected = radius_s_ii / 2
    assert np.isclose(result, expected)


def test_qian_2022_rg():
    """Test the qian_2022_rg function."""
    result = qian_2022_rg(1.5, 0.1)
    expected = (0.973 * (1.5 / 0.1) + 0.441) * 0.1
    assert np.isclose(result, expected)


def test_qian_2022_rg_df():
    """Test the qian_2022_rg_df function."""
    result = qian_2022_rg_df(2.5, 1.5, 0.1)
    expected = (0.882 * (2.5**0.223) * (1.5 / 0.1) + 0.387) * 0.1
    assert np.isclose(result, expected)


def test_qian_2022_rg_df_k0():
    """Test the qian_2022_rg_df_k0 function."""
    result = qian_2022_rg_df_k0(2.5, 1.2, 1.5, 0.1)
    expected = (
        0.777 * (2.5**0.479) * (1.2**0.000970) * (1.5 / 0.1)
        + 0.267 * 1.2
        + -0.0790
    ) * 0.1
    assert np.isclose(result, expected)


def test_qian_2022_rg_df_k0_a13():
    """Test the qian_2022_rg_df_k0_a13 function."""
    result = qian_2022_rg_df_k0_a13(2.5, 1.2, 0.5, 1.5, 0.1)
    expected = (
        0.876 * (2.5**0.363) * (1.2**-0.105) * (1.5 / 0.1)
        + 0.421 * 1.2
        + -0.0360 * 0.5
        + -0.227
    ) * 0.1
    assert np.isclose(result, expected)



"""Particle resolved method for coagulation.
"""

from typing import Tuple, Union
import numpy as np
from numpy.typing import NDArray
from scipy.interpolate import RectBivariateSpline  # type: ignore

from particula.dynamics.coagulation.super_droplet_method import (
    bin_particles,
    get_bin_pairs,
)


def interpolate_kernel(
    kernel: NDArray[np.float64],
    kernel_radius: NDArray[np.float64],
) -> RectBivariateSpline:
    """
    Create a 2D interpolation function for the coagulation kernel.

    Args:
        kernel (NDArray[np.float64]): Coagulation kernel.
        kernel_radius (NDArray[np.float64]): Radii corresponding to kernel
            bins.

    Returns:
        RectBivariateSpline: Interpolated kernel function.
    """
    return RectBivariateSpline(x=kernel_radius, y=kernel_radius, z=kernel)


def calculate_probabilities(
    kernel_values: Union[float, NDArray[np.float64]],
    time_step: float,
    events: int,
    tests: int,
    volume: float,
) -> Union[float, NDArray[np.float64]]:
    """
    Calculate coagulation probabilities based on kernel values and system
    parameters.

    Args:
        kernel_values (float): Interpolated kernel value for a particle pair.
        time_step (float): The time step over which coagulation occurs.
        events (int): Number of possible coagulation events.
        tests (int): Number of tests (or trials) for coagulation.
        volume (float): Volume of the system.

    Returns:
        float: Coagulation probability.
    """
    return kernel_values * time_step * events / (tests * volume)


def resolve_final_coagulation_state(
    small_indices: NDArray[np.int64],
    large_indices: NDArray[np.int64],
    particle_radius: NDArray[np.float64],
) -> Tuple[NDArray[np.int64], NDArray[np.int64]]:
    """
    Resolve the final state of particles that have undergone multiple
    coagulation events.

    Args:
        small_indices (NDArray[np.int64]): Indices of smaller particles.
        large_indices (NDArray[np.int64]): Indices of larger particles.
        particle_radius (NDArray[np.float64]): Radii of particles.

    Returns:
        Tuple[NDArray[np.int64], NDArray[np.int64]]: Updated small and large
        indices.
    """
    # Find common indices that appear in both small and large arrays
    commons, small_common, large_common = np.intersect1d(
        small_indices, large_indices, return_indices=True
    )
    # Sort particles by radius to resolve final states in the correct order
    sorted_args = np.argsort(particle_radius[commons])
    commons = commons[sorted_args]
    small_common = small_common[sorted_args]
    large_common = large_common[sorted_args]

    # Remap to the largest particle in common to resolve the final state
    for i, common in enumerate(commons):
        final_value = large_indices[small_common[i]]
        remap_index = np.flatnonzero(large_indices == common)
        large_indices[remap_index] = final_value

    return small_indices, large_indices


def particle_resolved_update_step(
    particle_radius: NDArray[np.float64],
    loss: NDArray[np.float64],
    gain: NDArray[np.float64],
    small_index: NDArray[np.int64],
    large_index: NDArray[np.int64],
) -> Tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.float64]]:
    """
    Update the particle radii and concentrations after coagulation events.

    Args:
        particle_radius (NDArray[float64]): Array of particle radii.
        small_index (NDArray[int64]): Indices corresponding to smaller
            particles.
        large_index (NDArray[int64]): Indices corresponding to larger
            particles.

    Returns:
        - Updated array of particle radii.
        - Updated array for the radii of particles that were lost.
        - Updated array for the radii of particles that were gained.
    """
    # Step 1: Calculate the summed volumes of the smaller and larger particles
    # The volumes are obtained by cubing the radii of the particles.
    sum_radii_cubed = np.power(
        particle_radius[small_index], 3, dtype=np.float64
    ) + np.power(particle_radius[large_index], 3, dtype=np.float64)

    # Step 2: Calculate the new radii formed by the coagulation events
    # The new radius is the cube root of the summed volumes.
    new_radii = np.cbrt(sum_radii_cubed)

    # Step 3: Save out the loss and gain of particles
    loss[small_index] = particle_radius[small_index]
    gain[large_index] = particle_radius[large_index]

    # Step 4: Remove the small particles as they coagulated to the larger ones
    particle_radius[small_index] = 0

    # Step 5: Increase the radii of the large particles to the new radii
    particle_radius[large_index] = new_radii

    return particle_radius, loss, gain


# pylint: disable=too-many-positional-arguments, too-many-arguments
# pylint: disable=too-many-locals
def particle_resolved_coagulation_step(
    particle_radius: NDArray[np.float64],
    kernel: NDArray[np.float64],
    kernel_radius: NDArray[np.float64],
    volume: float,
    time_step: float,
    random_generator: np.random.Generator,
) -> NDArray[np.int64]:
    """
    Perform a single step of particle coagulation, updating particle radii
    based on coagulation events.

    Args:
        particle_radius (NDArray[np.float64]): Array of particle radii.
        kernel (NDArray[np.float64]): Coagulation kernel as a 2D array where
            each element represents the probability of coagulation between
            particles of corresponding sizes.
        kernel_radius (NDArray[np.float64]): Array of radii corresponding to
            the kernel bins.
        volume (float): Volume of the system in which coagulation occurs.
        time_step (float): Time step over which coagulation is calculated.
        random_generator (np.random.Generator): Random number generator for
            stochastic processes.

    Returns:
        NDArray[np.int64]: Array of indices corresponding to the coagulation
            events, where each element is a pair of indices corresponding to
            the coagulating particles [loss, gain].
    """

    # Step 1: Bin the particles based on their radii into corresponding kernel
    # bins
    _, bin_indices = bin_particles(particle_radius, kernel_radius)

    # Step 2: Precompute unique bin pairs for efficient coagulation
    pair_indices = get_bin_pairs(bin_indices=bin_indices)

    # Step 3: Interpolate the coagulation kernel for efficient lookups during
    # the coagulation process
    interp_kernel = interpolate_kernel(kernel, kernel_radius)

    # Create output arrays to store the indices of small and large particles
    # involved in coagulation events
    small_index_total = np.array([], dtype=np.int64)
    large_index_total = np.array([], dtype=np.int64)

    # Step 4: Iterate over each bin pair to calculate potential coagulation
    # events
    for lower_bin, upper_bin in pair_indices:
        # Get indices of particles in the current bin and filter out any that
        # have already coagulated
        small_indices = np.flatnonzero(
            (bin_indices == lower_bin) & (particle_radius > 0)
        )
        small_indices = np.setdiff1d(small_indices, small_index_total)

        large_indices = np.flatnonzero(
            (bin_indices == upper_bin) & (particle_radius > 0)
        )

        # Skip to the next bin pair if there are no particles in one or
        # both bins
        if len(small_indices) == 0 or len(large_indices) == 0:
            continue

        # Step 5: Retrieve the maximum kernel value for the current bin pair
        kernel_values = interp_kernel.ev(  # type: ignore
            np.min(particle_radius[small_indices]),
            np.max(particle_radius[large_indices]),
        )

        # Step 6: Calculate the number of possible coagulation events
        # between small and large particles
        events = len(small_indices) * len(large_indices)
        if lower_bin == upper_bin:
            events = len(small_indices) * (len(large_indices) - 1) / 2
        events = int(np.ceil(events))

        # Step 7: Determine the number of coagulation tests to run based
        # on kernel value and system parameters
        tests = int(np.ceil(kernel_values * time_step * events / volume))

        if tests == 0 or events == 0:
            continue

        # Step 8: Randomly select small and large particle pairs for
        # coagulation tests
        replace_in_pool = tests > len(small_indices)
        small_index = random_generator.choice(  # type: ignore
            small_indices, size=tests, replace=bool(replace_in_pool)
        )
        large_index = random_generator.choice(large_indices, tests)

        # Step 9: Calculate the kernel value for the selected particle pairs
        kernel_value = interp_kernel.ev(  # type: ignore
            particle_radius[small_index], particle_radius[large_index]
        )

        # Handle diagonal elements if necessary (for single pair coagulation)
        if kernel_value.ndim > 1:
            kernel_value = np.diagonal(kernel_value)

        # Step 10: Calculate coagulation probabilities for each selected pair
        coagulation_probabilities = calculate_probabilities(
            kernel_value, time_step, events, tests, volume
        )

        # Step 11: Determine which coagulation events occur based on
        # random uniform sampling
        valid_indices = np.flatnonzero(  # type: ignore
            random_generator.uniform(size=tests) < coagulation_probabilities
        )

        # Step 12: Ensure each small particle only coagulates with one large
        # particle
        _, unique_index = np.unique(
            small_index[valid_indices], return_index=True
        )
        # Valid and unique indices are selected for coagulation,
        # non-unique indices should happen rarely.
        small_index = small_index[valid_indices][unique_index]
        large_index = large_index[valid_indices][unique_index]

        # Step 13: Save the coagulation events
        small_index_total = np.append(small_index_total, small_index)
        large_index_total = np.append(large_index_total, large_index)

    # Step 14: Resolve any series of coagulation events that involve the same
    # particles multiple times
    small_index_total, large_index_total = resolve_final_coagulation_state(
        small_index_total, large_index_total, particle_radius
    )

    # Step 15: Combine small and large indices into pairs representing the
    # loss and gain events
    loss_gain_index = np.column_stack([small_index_total, large_index_total])

    return loss_gain_index



"""Tests for the aerodynamic_mobility module."""

from contextlib import nullcontext
import numpy as np
import pytest
from particula.particles.properties import particle_aerodynamic_mobility


def test_particle_aerodynamic_mobility_single_value():
    """
    Verify that the particle_aerodynamic_mobility function calculates the
    correct aerodynamic mobility value for a single particle.
    """
    radius = 0.00005  # 50 micrometers
    slip_correction_factor = 1.1
    dynamic_viscosity = 0.0000181  # Pa.s for air at room temperature
    expected_mobility = slip_correction_factor / (
        6 * np.pi * dynamic_viscosity * radius
    )
    actual_mobility = particle_aerodynamic_mobility(
        radius, slip_correction_factor, dynamic_viscosity
    )
    assert np.isclose(
        actual_mobility, expected_mobility
    ), "The value does not match."


def test_particle_aerodynamic_mobility_array_input():
    """
    Test that the particle_aerodynamic_mobility function handles numpy array
    inputs correctly.
    """
    radius = np.array([0.00005, 0.00007])  # Array of radii
    slip_correction_factor = np.array([1.1, 1.2])
    dynamic_viscosity = 0.0000181
    expected_mobility = slip_correction_factor / (
        6 * np.pi * dynamic_viscosity * radius
    )
    actual_mobility = particle_aerodynamic_mobility(
        radius, slip_correction_factor, dynamic_viscosity
    )
    assert np.allclose(actual_mobility, expected_mobility)


def test_particle_aerodynamic_mobility_type_error():
    """
    Test that the particle_aerodynamic_mobility function raises a TypeError
    with incorrect input types.
    """
    with pytest.raises(TypeError):
        particle_aerodynamic_mobility(
            "not a number", "also not a number", "still not a number"
        )


@pytest.mark.parametrize(
    "radius", [0, -1e-6, 1e-12, 1e-3, float("inf"), float("nan")]
)
def test_particle_aerodynamic_mobility_extreme_values(radius: float) -> None:
    """
    Verify that the particle_aerodynamic_mobility function handles extreme
    values correctly.
    - radius : The radius of the particle
    """
    slip_correction_factor = 1.1
    dynamic_viscosity = 0.0000181  # Pa.s for air at room temperature
    with (
        pytest.raises(ValueError)
        if (radius <= 0 or np.isnan(radius) or np.isinf(radius))
        else nullcontext()
    ):
        actual_mobility = particle_aerodynamic_mobility(
            radius, slip_correction_factor, dynamic_viscosity
        )
        if radius > 0:
            expected_mobility = slip_correction_factor / (
                6 * np.pi * dynamic_viscosity * radius
            )
            assert np.isclose(
                actual_mobility, expected_mobility
            ), f"The value does not match for radius {radius}."



"""Builder class for Surface strategies, for how to account for
surface tension in the calculation of the Kelvin effect.

This builds the strategy and checks that the required parameters are set,
and converts the units of the parameters if necessary.
"""

import logging
from particula.abc_builder import (
    BuilderABC,
)
from particula.builder_mixin import (
    BuilderDensityMixin,
    BuilderSurfaceTensionMixin,
    BuilderMolarMassMixin,
)
from particula.particles.surface_strategies import (
    SurfaceStrategyMass,
    SurfaceStrategyMolar,
    SurfaceStrategyVolume,
)

logger = logging.getLogger("particula")


class SurfaceStrategyMolarBuilder(
    BuilderABC,
    BuilderDensityMixin,
    BuilderSurfaceTensionMixin,
    BuilderMolarMassMixin,
):
    """Builder class for SurfaceStrategyMolar objects.

    Methods:
        set_surface_tension(surface_tension, surface_tension_units): Set the
            surface tension of the particle in N/m. Default units are 'N/m'.
        set_density(density, density_units): Set the density of the particle in
            kg/m^3. Default units are 'kg/m^3'.
        set_molar_mass(molar_mass, molar_mass_units): Set the molar mass of the
            particle in kg/mol. Default units are 'kg/mol'.
        set_parameters(params): Set the parameters of the SurfaceStrategyMolar
            object from a dictionary including optional units.
        build(): Validate and return the SurfaceStrategyMolar object.
    """

    def __init__(self):
        required_parameters = ["surface_tension", "density", "molar_mass"]
        BuilderABC.__init__(self, required_parameters)
        BuilderSurfaceTensionMixin.__init__(self)
        BuilderDensityMixin.__init__(self)
        BuilderMolarMassMixin.__init__(self)

    def build(self) -> SurfaceStrategyMolar:
        """Validate and return the SurfaceStrategyMass object.

        Returns:
            SurfaceStrategyMolar: Instance of the SurfaceStrategyMolar object.
        """
        self.pre_build_check()
        return SurfaceStrategyMolar(
            surface_tension=self.surface_tension,  # type: ignore
            density=self.density,  # type: ignore
            molar_mass=self.molar_mass,  # type: ignore
        )


class SurfaceStrategyMassBuilder(
    BuilderABC, BuilderSurfaceTensionMixin, BuilderDensityMixin
):
    """Builder class for SurfaceStrategyMass objects.

    Methods:
        set_surface_tension(surface_tension, surface_tension_units): Set the
            surface tension of the particle in N/m. Default units are 'N/m'.
        set_density(density, density_units): Set the density of the particle in
            kg/m^3. Default units are 'kg/m^3'.
        set_parameters(params): Set the parameters of the SurfaceStrategyMass
            object from a dictionary including optional units.
        build(): Validate and return the SurfaceStrategyMass object.
    """

    def __init__(self):
        required_parameters = ["surface_tension", "density"]
        BuilderABC.__init__(self, required_parameters)
        BuilderSurfaceTensionMixin.__init__(self)
        BuilderDensityMixin.__init__(self)

    def build(self) -> SurfaceStrategyMass:
        """Validate and return the SurfaceStrategyMass object.

        Returns:
            SurfaceStrategyMass: Instance of the SurfaceStrategyMass object.
        """
        self.pre_build_check()
        return SurfaceStrategyMass(
            surface_tension=self.surface_tension,  # type: ignore
            density=self.density,  # type: ignore
        )


class SurfaceStrategyVolumeBuilder(
    BuilderABC, BuilderSurfaceTensionMixin, BuilderDensityMixin
):
    """Builder class for SurfaceStrategyVolume objects.

    Methods:
        set_surface_tension(surface_tension, surface_tension_units): Set the
            surface tension of the particle in N/m. Default units are 'N/m'.
        set_density(density, density_units): Set the density of the particle in
            kg/m^3. Default units are 'kg/m^3'.
        set_parameters(params): Set the parameters of the SurfaceStrategyVolume
            object from a dictionary including optional units.
        build(): Validate and return the SurfaceStrategyVolume object.
    """

    def __init__(self):
        required_parameters = ["surface_tension", "density"]
        BuilderABC.__init__(self, required_parameters)
        BuilderSurfaceTensionMixin.__init__(self)
        BuilderDensityMixin.__init__(self)

    def build(self) -> SurfaceStrategyVolume:
        """Validate and return the SurfaceStrategyVolume object.

        Returns:
            SurfaceStrategyVolume: Instance of the SurfaceStrategyVolume
                object.
        """
        self.pre_build_check()
        return SurfaceStrategyVolume(
            surface_tension=self.surface_tension,  # type: ignore
            density=self.density,  # type: ignore
        )



"""Test rate calculations for coagulation dynamics.

tested for evaluation not accuracy of the calculations."""

import numpy as np
from particula.dynamics.coagulation import rate

# Define constants for test data
# Define constants for test data
CONCENTRATION = np.array([1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0])
KERNEL = np.array(
    [
        [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7],
        [0.2, 0.4, 0.6, 0.8, 1.0, 1.2, 1.4],
        [0.3, 0.6, 0.9, 1.2, 1.5, 1.8, 2.1],
        [0.4, 0.8, 1.2, 1.6, 2.0, 2.4, 2.8],
        [0.5, 1.0, 1.5, 2.0, 2.5, 3.0, 3.5],
        [0.6, 1.2, 1.8, 2.4, 3.0, 3.6, 4.2],
        [0.7, 1.4, 2.1, 2.8, 3.5, 4.2, 4.9],
    ]
)
RADIUS = np.array([0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7])


def test_discrete_loss():
    """
    Test the discrete_loss function with predefined concentration and kernel
    values.
    """
    result = rate.discrete_loss(CONCENTRATION, KERNEL)
    expected = np.array([14.0, 56.0, 126.0, 224.0, 350.0, 504.0, 686.0])
    np.testing.assert_almost_equal(result, expected, decimal=6)


def test_discrete_gain():
    """
    Test the discrete_gain function with predefined concentration and kernel
    values.
    """
    result = rate.discrete_gain(RADIUS, CONCENTRATION, KERNEL)
    expected = np.array(
        [
            8.88984299e-02,
            8.07430616e-01,
            4.71791309e00,
            1.83365875e01,
            5.45636475e01,
            1.33885983e02,
            2.86922649e02,
        ]
    )
    np.testing.assert_almost_equal(result, expected, decimal=6)


def test_continuous_loss():
    """
    Test the continuous_loss function with predefined radius, concentration,
    and kernel values.
    """
    result = rate.continuous_loss(RADIUS, CONCENTRATION, KERNEL)
    expected = np.array([1.15, 4.6, 10.35, 18.4, 28.75, 41.4, 56.35])
    np.testing.assert_almost_equal(result, expected, decimal=6)


def test_continuous_gain():
    """
    Test the continuous_gain function with predefined radius, concentration,
    and kernel values.
    """
    result = rate.continuous_gain(RADIUS, CONCENTRATION, KERNEL)
    expected = np.array(
        [
            8.88984299e-03,
            8.07430616e-02,
            4.71791309e-01,
            1.83365875e00,
            5.45636475e00,
            1.33885983e01,
            2.86922649e01,
        ]
    )
    np.testing.assert_almost_equal(result, expected, decimal=6)



"""Module for aerodynamic size and shape factor of a particle in a fluid.
"""

from typing import Union
from numpy.typing import NDArray
import numpy as np


AERODYNAMIC_SHAPE_FACTOR_DICT = {
    "sphere": 1.0,
    "cube": 1.08,
    "cylinder_avg_aspect_2": 1.10,
    "cylinder_avg_aspect_5": 1.35,
    "cylinder_avg_aspect_10": 1.68,
    "spheres_cluster_3": 1.15,
    "spheres_cluster_4": 1.17,
    "bituminous_coal": 1.08,
    "quartz": 1.36,
    "sand": 1.57,
    "talc": 1.88,
}


# pylint: disable=too-many-positional-arguments, too-many-arguments
def particle_aerodynamic_length(
    physical_length: Union[float, NDArray[np.float64]],
    physical_slip_correction_factor: Union[float, NDArray[np.float64]],
    aerodynamic_slip_correction_factor: Union[float, NDArray[np.float64]],
    density: Union[float, NDArray[np.float64]],
    reference_density: float = 1000,
    aerodynamic_shape_factor: float = 1.0,
) -> Union[float, NDArray[np.float64]]:
    """
    Calculate the aerodynamic length scale of a particle.

    The aerodynamic size is used to compare the aerodynamic properties of
    particles with their physical properties, particularly when interpreting
    aerodynamic particle sizer measurements. This is either the aerodynamic
    diameter or radius, depending on length scale input.

    Args:
        - physical_length : Physical length scale of the particle (m).
        - physical_slip_correction_factor : Slip correction factor for the
             particle's physical radius in the fluid (dimensionless).
        - aerodynamic_slip_correction_factor : Slip correction factor for the
             particle's aerodynamic radius in the fluid (dimensionless).
        - density : Density of the particle (kg/m^3).
        - reference_density : Reference density for the particle, typically
             the density of water (1000 kg/m^3 by default).
        - aerodynamic_shape_factor : Shape factor of the particle, accounting
             for non-sphericity (dimensionless, default is 1.0 for spherical
             particles).

    Returns:
        Aerodynamic radius of the particle (m).

    Example:
        ``` py title="Example"
        aerodynamic_length = particle_aerodynamic_length(
            physical_length=0.00005,
            physical_slip_correction_factor=1.1,
            aerodynamic_slip_correction_factor=1.0,
            density=1000,
            reference_density=1000,
            aerodynamic_shape_factor=1.0,
        )
        ```

    References:
        - https://en.wikipedia.org/wiki/Aerosol#Aerodynamic_diameter
        - Hinds, W.C. (1998) Aerosol Technology: Properties, behavior, and
            measurement of airborne particles. Wiley-Interscience, New York.
            pp 51-53, section 3.6.
    """
    return physical_length * np.sqrt(
        (physical_slip_correction_factor / aerodynamic_slip_correction_factor)
        * (density / (reference_density * aerodynamic_shape_factor))
    )


def get_aerodynamic_shape_factor(shape_key: str) -> float:
    """Retrieve the aerodynamic shape factor for a given particle shape.

    Args:
        - shape_key : The shape of the particle as a string. Options include:
            sphere, cube, cylinder_avg_aspect_2, cylinder_avg_aspect_5,
            cylinder_avg_aspect_10, spheres_cluster_3, spheres_cluster_4,
            bituminous_coal, quartz, sand, talc

    Returns:
        The shape factor of the particle as a float.

    Example:
        ``` py title="Example"
        shape_factor = get_aerodynamic_shape_factor("sand")
        # shape_factor = 1.57
        ```

    Raises:
        - ValueError : If the shape is not found in the predefined shape
            factor dictionary.
    """
    shape_key = shape_key.strip().lower()  # Clean up the input

    # Retrieve the shape factor from the dictionary, or raise an error
    try:
        return AERODYNAMIC_SHAPE_FACTOR_DICT[shape_key]
    except KeyError as exc:
        raise ValueError(
            f"The shape factor for the shape '{shape_key}' is not available."
        ) from exc



"""Strategies for surface effects on particles.

Should add the organic film strategy in the future.
"""

from abc import ABC, abstractmethod
from typing import Union
from numpy.typing import NDArray
import numpy as np

from particula.util.convert import (
    mass_concentration_to_mole_fraction,
    mass_concentration_to_volume_fraction,
)
from particula.particles.properties import kelvin_radius, kelvin_term


class SurfaceStrategy(ABC):
    """ABC class for Surface Strategies.

    Abstract class for implementing strategies to calculate surface tension
    and the Kelvin effect for species in particulate phases.

    Methods:
        effective_surface_tension: Calculate the effective surface tension of
            species based on their concentration.
        effective_density: Calculate the effective density of species based on
            their concentration.
        get_name: Return the type of the surface strategy.
        kelvin_radius: Calculate the Kelvin radius which determines the
            curvature effect on vapor pressure.
        kelvin_term: Calculate the Kelvin term, which quantifies the effect of
            particle curvature on vapor pressure.
    """

    @abstractmethod
    def effective_surface_tension(
        self, mass_concentration: Union[float, NDArray[np.float64]]
    ) -> float:
        """Calculate the effective surface tension of the species mixture.

        Args:
            mass_concentration: Concentration of the species [kg/m^3].

        Returns:
            float or NDArray[float]: Effective surface tension [N/m].
        """

    @abstractmethod
    def effective_density(
        self, mass_concentration: Union[float, NDArray[np.float64]]
    ) -> float:
        """Calculate the effective density of the species mixture.

        Args:
            mass_concentration: Concentration of the species [kg/m^3].

        Returns:
            float or NDArray[float]: Effective density of the species [kg/m^3].
        """

    def get_name(self) -> str:
        """Return the type of the surface strategy."""
        return self.__class__.__name__

    def kelvin_radius(
        self,
        molar_mass: Union[float, NDArray[np.float64]],
        mass_concentration: Union[float, NDArray[np.float64]],
        temperature: float,
    ) -> Union[float, NDArray[np.float64]]:
        """Calculate the Kelvin radius which determines the curvature effect.

        The kelvin radius is molecule specific and depends on the surface
        tension, molar mass, density, and temperature of the system. It is
        used to calculate the Kelvin term, which quantifies the effect of
        particle curvature on vapor pressure.

        Args:
            surface_tension: Surface tension of the mixture [N/m].
            molar_mass: Molar mass of the species [kg/mol].
            mass_concentration: Concentration of the species [kg/m^3].
            temperature: Temperature of the system [K].

        Returns:
            float or NDArray[float]: Kelvin radius [m].

        References:
        - Based on Neil Donahue's approach to the Kelvin equation:
        r = 2 * surface_tension * molar_mass / (R * T * density)
        [Kelvin Wikipedia](https://en.wikipedia.org/wiki/Kelvin_equation)
        """
        return kelvin_radius(
            self.effective_surface_tension(mass_concentration),
            self.effective_density(mass_concentration),
            molar_mass,
            temperature,
        )

    def kelvin_term(
        self,
        radius: Union[float, NDArray[np.float64]],
        molar_mass: Union[float, NDArray[np.float64]],
        mass_concentration: Union[float, NDArray[np.float64]],
        temperature: float,
    ) -> Union[float, NDArray[np.float64]]:
        """Calculate the Kelvin term, which multiplies the vapor pressure.

        The Kelvin term is used to adjust the vapor pressure of a species due
        to the curvature of the particle.

        Args:
            radius: Radius of the particle [m].
            molar_mass: Molar mass of the species a [kg/mol].
            mass_concentration: Concentration of the species [kg/m^3].
            temperature: Temperature of the system [K].

        Returns:
            float or NDArray[float]: The exponential factor adjusting vapor
                pressure due to curvature.

        References:
        Based on Neil Donahue's approach to the Kelvin equation:
        exp(kelvin_radius / particle_radius)
        [Kelvin Eq Wikipedia](https://en.wikipedia.org/wiki/Kelvin_equation)
        """
        return kelvin_term(
            radius,
            self.kelvin_radius(molar_mass, mass_concentration, temperature),
        )


# Surface mixing strategies
class SurfaceStrategyMolar(SurfaceStrategy):
    """Surface tension and density, based on mole fraction weighted values.

    Arguments:
        surface_tension: Surface tension of the species [N/m]. If a single
            value is provided, it will be used for all species.
        density: Density of the species [kg/m^3]. If a single value is
            provided, it will be used for all species.
        molar_mass: Molar mass of the species [kg/mol]. If a single value is
            provided, it will be used for all species.

    References:
        [Mole Fractions](https://en.wikipedia.org/wiki/Mole_fraction)
    """

    def __init__(
        self,
        surface_tension: Union[float, NDArray[np.float64]] = 0.072,  # water
        density: Union[float, NDArray[np.float64]] = 1000,  # water
        molar_mass: Union[float, NDArray[np.float64]] = 0.01815,  # water
    ):
        self.surface_tension = surface_tension
        self.density = density
        self.molar_mass = molar_mass

    def effective_surface_tension(
        self, mass_concentration: Union[float, NDArray[np.float64]]
    ) -> float:
        if isinstance(self.surface_tension, float):
            return self.surface_tension
        return np.sum(
            self.surface_tension
            * mass_concentration_to_mole_fraction(
                mass_concentration, self.molar_mass  # type: ignore
            ),
            dtype=np.float64,
        )

    def effective_density(
        self, mass_concentration: Union[float, NDArray[np.float64]]
    ) -> float:
        if isinstance(self.density, float):
            return self.density
        return np.sum(
            self.density
            * mass_concentration_to_mole_fraction(
                mass_concentration, self.molar_mass  # type: ignore
            ),
            dtype=np.float64,
        )


class SurfaceStrategyMass(SurfaceStrategy):
    """Surface tension and density, based on mass fraction weighted values.

    Args:
        surface_tension: Surface tension of the species [N/m]. If a single
            value is provided, it will be used for all species.
        density: Density of the species [kg/m^3]. If a single value is
            provided, it will be used for all species.

    References:
    [Mass Fractions](https://en.wikipedia.org/wiki/Mass_fraction_(chemistry))
    """

    def __init__(
        self,
        surface_tension: Union[float, NDArray[np.float64]] = 0.072,  # water
        density: Union[float, NDArray[np.float64]] = 1000,  # water
    ):
        self.surface_tension = surface_tension
        self.density = density

    def effective_surface_tension(
        self, mass_concentration: Union[float, NDArray[np.float64]]
    ) -> float:
        if isinstance(self.surface_tension, float):
            return self.surface_tension
        return np.sum(
            self.surface_tension
            * mass_concentration
            / np.sum(mass_concentration),
            dtype=np.float64,
        )

    def effective_density(
        self, mass_concentration: Union[float, NDArray[np.float64]]
    ) -> float:
        if isinstance(self.density, float):
            return self.density
        return np.sum(
            self.density * mass_concentration / np.sum(mass_concentration),
            dtype=np.float64,
        )


class SurfaceStrategyVolume(SurfaceStrategy):
    """Surface tension and density, based on volume fraction weighted values.

    Args:
        surface_tension: Surface tension of the species [N/m]. If a single
            value is provided, it will be used for all species.
        density: Density of the species [kg/m^3]. If a single value is
            provided, it will be used for all species.

    References:
    [Volume Fractions](https://en.wikipedia.org/wiki/Volume_fraction)
    """

    def __init__(
        self,
        surface_tension: Union[float, NDArray[np.float64]] = 0.072,  # water
        density: Union[float, NDArray[np.float64]] = 1000,  # water
    ):
        self.surface_tension = surface_tension
        self.density = density

    def effective_surface_tension(
        self, mass_concentration: Union[float, NDArray[np.float64]]
    ) -> float:
        if isinstance(self.surface_tension, float):
            return self.surface_tension
        return np.sum(
            self.surface_tension
            * mass_concentration_to_volume_fraction(
                mass_concentration, self.density  # type: ignore
            ),
            dtype=np.float64,
        )

    def effective_density(
        self, mass_concentration: Union[float, NDArray[np.float64]]
    ) -> float:
        if isinstance(self.density, float):
            return self.density
        return np.sum(
            self.density
            * mass_concentration_to_volume_fraction(
                mass_concentration, self.density  # type: ignore
            ),
            dtype=np.float64,
        )



"""Tests for the calculate_partial_pressure function."""

import numpy as np
import pytest
from particula.util.constants import GAS_CONSTANT

from particula.gas.properties.pressure_function import (
    calculate_partial_pressure,
)


def test_calculate_partial_pressure_scalar():
    """Test the calculate_partial_pressure function with scalar values."""
    # Test the function with scalar values
    concentration = 1.0  # kg/m^3
    molar_mass = 0.029  # kg/mol (approx. for air)
    temperature = 298  # K (approx. 25 degrees Celsius)
    expected_pressure = (
        concentration * GAS_CONSTANT * temperature
    ) / molar_mass
    assert calculate_partial_pressure(
        concentration, molar_mass, temperature
    ) == pytest.approx(expected_pressure)


def test_calculate_partial_pressure_array():
    """Test the calculate_partial_pressure function with NumPy arrays."""
    # Test the function with NumPy arrays
    concentration = np.array([1.0, 2.0])  # kg/m^3
    molar_mass = np.array([0.029, 0.032])  # kg/mol (approx. for air and O2)
    temperature = np.array([298, 310])  # K
    expected_pressure = (
        concentration * GAS_CONSTANT * temperature
    ) / molar_mass
    np.testing.assert_array_almost_equal(
        calculate_partial_pressure(concentration, molar_mass, temperature),
        expected_pressure,
    )


def test_calculate_partial_pressure_edge_cases():
    """Test the calculate_partial_pressure function with edge case values."""
    test_cases = [
        (1e-6, 0.029, 298),  # Very low concentration
        (1e6, 0.029, 298),  # Very high concentration
        (1.0, 0.001, 298),  # Very low molar mass
        (1.0, 1.0, 298),    # High molar mass
        (1.0, 0.029, 1),    # Very low temperature
        (1.0, 0.029, 1e4),  # Very high temperature
    ]
    for concentration, molar_mass, temperature in test_cases:
        expected_pressure = (
            concentration * GAS_CONSTANT * temperature
        ) / molar_mass
        assert calculate_partial_pressure(
            concentration, molar_mass, temperature
        ) == pytest.approx(expected_pressure)



"""Runnable process that are particle centric.
Includes, condensation (and evaporation), coagulation, and deposition.
"""

from typing import Any
import numpy as np

# Particula imports
from particula.runnable import Runnable
from particula.aerosol import Aerosol
from particula.dynamics.condensation.condensation_strategies import (
    CondensationStrategy,
)
from particula.dynamics.coagulation.strategy import CoagulationStrategy


class MassCondensation(Runnable):
    """
    A class for running a mass condensation process.

    Args:
        condensation_strategy (CondensationStrategy): The condensation strategy
            to use.

    Methods:
        execute: Execute the mass condensation process.
        rate: Calculate the rate of mass condensation for each particle due to
            each condensable gas species.
    """

    def __init__(self, condensation_strategy: CondensationStrategy):
        self.condensation_strategy = condensation_strategy

    def execute(
        self, aerosol: Aerosol, time_step: float, sub_steps: int = 1
    ) -> Aerosol:
        """
        Execute the mass condensation process.

        Args:
            aerosol (Aerosol): The aerosol instance to modify.
        """
        # loop over gas species
        for gas_species in aerosol.iterate_gas():
            # check if gas species is condensable
            if not gas_species.condensable:
                continue
            # loop over particles to apply condensation
            for particle in aerosol.iterate_particle():
                for _ in range(sub_steps):
                    # calculate the condensation step for strategy
                    particle, gas_species = self.condensation_strategy.step(
                        particle=particle,
                        gas_species=gas_species,
                        temperature=aerosol.atmosphere.temperature,
                        pressure=aerosol.atmosphere.total_pressure,
                        time_step=time_step / sub_steps,
                    )
        return aerosol

    def rate(self, aerosol: Aerosol) -> Any:
        """
        Calculate the rate of mass condensation for each particle due to
        each condensable gas species.

        Args:
            aerosol (Aerosol): The aerosol instance to modify.

        Returns:
            np.ndarray: An array of condensation rates for each particle.
        """
        rates = np.array([], dtype=np.float64)
        # Loop over gas species in the aerosol
        for gas_species in aerosol.iterate_gas():
            # Check if the gas species is condensable
            if not gas_species.condensable:
                continue
            # Loop over particles to apply condensation
            for particle in aerosol.iterate_particle():
                # print(gas_species.name)
                # Calculate the rate of condensation
                mass_rate = self.condensation_strategy.rate(
                    particle=particle,
                    gas_species=gas_species,
                    temperature=aerosol.atmosphere.temperature,
                    pressure=aerosol.atmosphere.total_pressure,
                )
                rates = np.append(rates, mass_rate)
        return rates


class Coagulation(Runnable):
    """
    A class for running a coagulation strategy.

    Args:
        coagulation_strategy (CoagulationStrategy): The coagulation strategy to
            use.

    Methods:
        execute: Execute the coagulation process.
        rate: Calculate the rate of coagulation for each particle.
    """

    def __init__(self, coagulation_strategy: CoagulationStrategy):
        self.coagulation_strategy = coagulation_strategy

    def execute(
        self, aerosol: Aerosol, time_step: float, sub_steps: int = 1
    ) -> Aerosol:
        """
        Execute the coagulation process.

        Args:
            aerosol (Aerosol): The aerosol instance to modify.
        """
        # Loop over particles
        for particle in aerosol.iterate_particle():
            for _ in range(sub_steps):
                # Calculate the coagulation step for the particle
                particle = self.coagulation_strategy.step(
                    particle=particle,
                    temperature=aerosol.atmosphere.temperature,
                    pressure=aerosol.atmosphere.total_pressure,
                    time_step=time_step / sub_steps,
                )
        return aerosol

    def rate(self, aerosol: Aerosol) -> Any:
        """
        Calculate the rate of coagulation for each particle.

        Args:
            aerosol (Aerosol): The aerosol instance to modify.

        Returns:
            np.ndarray: An array of coagulation rates for each particle.
        """
        rates = np.array([], dtype=np.float64)
        # Loop over particles
        for particle in aerosol.iterate_particle():
            # Calculate the net coagulation rate for the particle
            net_rate = self.coagulation_strategy.net_rate(
                particle=particle,
                temperature=aerosol.atmosphere.temperature,
                pressure=aerosol.atmosphere.total_pressure,
            )
            rates = np.append(rates, net_rate)
        return rates



"""Tests for the wall loss rate dynamics module."""

import numpy as np
from particula.dynamics.wall_loss import (
    spherical_wall_loss_rate,
    rectangle_wall_loss_rate,
)


def test_spherical_wall_loss_rate():
    """Test the spherical wall loss rate function."""
    # Test case 1
    wall_eddy_diffusivity = 0.1
    particle_radius = 1e-6
    particle_density = 1000.0
    particle_concentration = 100.0 * 1e6
    temperature = 298.0
    pressure = 101325.0
    chamber_radius = 10.0

    result = spherical_wall_loss_rate(
        wall_eddy_diffusivity,
        particle_radius,
        particle_density,
        particle_concentration,
        temperature,
        pressure,
        chamber_radius,
    )

    expected_loss_rate = -964.512508874514
    assert np.isclose(result, expected_loss_rate)

    # Test case 2
    wall_eddy_diffusivity = 0.05
    particle_radius = np.array([1e-9, 2e-6, 3e-3])
    particle_density = np.array([1000.0, 2000.0, 3000.0])
    particle_concentration = np.array([100.0, 200.0, 300.0]) * 1e6
    temperature = 273.15
    pressure = 100000.0
    chamber_radius = 5.0

    result = spherical_wall_loss_rate(
        wall_eddy_diffusivity,
        particle_radius,
        particle_density,
        particle_concentration,
        temperature,
        pressure,
        chamber_radius,
    )
    expected_loss_rate = np.array(
        [-9.22955735e03, -3.16327419e04, -1.54304331e11]
    )
    assert np.allclose(result, expected_loss_rate)


def test_rectangle_wall_loss_rate():
    """Test the rectangle wall loss rate function."""
    # Test single float input
    results = rectangle_wall_loss_rate(
        wall_eddy_diffusivity=0.1,
        particle_radius=1e-6,
        particle_density=1000.0,
        particle_concentration=100.0 * 1e6,
        temperature=298.0,
        pressure=101325.0,
        chamber_dimensions=(10.0, 10.0, 10.0),
    )

    expected_loss_rate = -1314.6750692789078
    assert np.isclose(results, expected_loss_rate)

    # Test single array input
    results = rectangle_wall_loss_rate(
        wall_eddy_diffusivity=0.05,
        particle_radius=np.array([1e-9, 2e-6, 3e-3]),
        particle_density=np.array([1000.0, 2000.0, 3000.0]),
        particle_concentration=np.array([100.0, 200.0, 300.0]) * 1e6,
        temperature=273.15,
        pressure=100000.0,
        chamber_dimensions=(5.0, 5.0, 5.0),
    )

    expected_loss_rate = np.array(
        [-1.84776108e04, -4.22330075e04, -2.05739107e11]
    )
    assert np.allclose(results, expected_loss_rate)



"""Module for friction factor of a particle in a fluid.

Zhang, C., Thajudeen, T., Larriba, C., Schwartzentruber, T. E., &
Hogan, C. J. (2012). Determination of the Scalar Friction Factor for
Nonspherical Particles and Aggregates Across the Entire Knudsen Number Range
by Direct Simulation Monte Carlo (DSMC). Aerosol Science and Technology,
46(10), 1065-1078. https://doi.org/10.1080/02786826.2012.690543
"""

from typing import Union
from numpy.typing import NDArray
import numpy as np


def friction_factor(
    radius: Union[float, NDArray[np.float64]],
    dynamic_viscosity: float,
    slip_correction: Union[float, NDArray[np.float64]],
):
    """
    Returns a particle's friction factor. Property of the particle's size and
    surrounding medium. Multiplying the friction factor by the fluid velocity
    yields the drag force on the particle.

    Args:
    -----
    - radius : The radius of the particle [m].
    - dynamic_viscosity : The dynamic viscosity of the fluid [Pa s].
    - slip_correction : The slip correction factor [unitless].

    Returns:
    --------
    The friction factor of the particle [N s/m].

    References:
    ----------
    It is best thought of as an inverse of mobility or the ratio between
    thermal energy and diffusion coefficient. The modified Stoke's diffusion
    coefficient is defined as
    kT / (6 * np.pi * dyn_vis_air * radius / slip_corr)
    and thus the friction factor can be defined as
    (6 * np.pi * dyn_vis_air * radius / slip_corr).

    In the continuum limit (Kn -> 0; Cc -> 1):
    6 * np.pi * dyn_vis_air * radius

    In the kinetic limit (Kn -> inf):
    8.39 * (dyn_vis_air/mfp_air) * const * radius**2

    Zhang, C., Thajudeen, T., Larriba, C., Schwartzentruber, T. E., &
    Hogan, C. J. (2012). Determination of the Scalar Friction Factor for
    Nonspherical Particles and Aggregates Across the Entire Knudsen Number
    Range by Direct Simulation Monte Carlo (DSMC). Aerosol Science and
    Technology, 46(10), 1065-1078. https://doi.org/10.1080/02786826.2012.690543
    """

    return 6 * np.pi * dynamic_viscosity * radius / slip_correction



"""
Velocity correlation f2 test module.
"""

import pytest
import numpy as np
from particula.dynamics.coagulation.turbulent_dns_kernel.velocity_correlation_f2_ao2008 import (
    get_f2_longitudinal_velocity_correlation,
)


def test_compute_f2_longitudinal_velocity_correlation_scalar():
    """
    Test compute_f2_longitudinal_velocity_correlation with scalar inputs.
    """
    collisional_radius = 0.1  # [m]
    taylor_microscale = 0.05  # [m]
    eulerian_integral_length = 1.0  # [m]

    # Compute expected values using explicit computation
    beta = (np.sqrt(2) * taylor_microscale) / eulerian_integral_length
    sqrt_term_t = np.sqrt(1 - 2 * beta**2)
    denominator = 2 * sqrt_term_t

    exp_term_1 = np.exp(
        -2
        * collisional_radius
        / ((1 + sqrt_term_t) * eulerian_integral_length)
    )
    exp_term_2 = np.exp(
        -2
        * collisional_radius
        / ((1 - sqrt_term_t) * eulerian_integral_length)
    )

    expected = (1 / denominator) * (
        (1 + sqrt_term_t) * exp_term_1 - (1 - sqrt_term_t) * exp_term_2
    )

    result = get_f2_longitudinal_velocity_correlation(
        collisional_radius, taylor_microscale, eulerian_integral_length
    )

    assert np.isclose(
        result, expected, atol=1e-10
    ), f"Expected {expected}, but got {result}"


def test_compute_f2_longitudinal_velocity_correlation_array():
    """
    Test compute_f2_longitudinal_velocity_correlation with NumPy array inputs.
    """
    collisional_radius = np.array([0.1, 0.5, 1.0])  # [m]
    taylor_microscale = 0.05  # [m]
    eulerian_integral_length = 1.0  # [m]

    beta = (np.sqrt(2) * taylor_microscale) / eulerian_integral_length
    sqrt_term = np.sqrt(1 - 2 * beta**2)
    denominator = 2 * sqrt_term

    exp_term_1a = np.exp(
        -2 * collisional_radius / ((1 + sqrt_term) * eulerian_integral_length)
    )
    exp_term_2b = np.exp(
        -2 * collisional_radius / ((1 - sqrt_term) * eulerian_integral_length)
    )

    expected = (1 / denominator) * (
        (1 + sqrt_term) * exp_term_1a - (1 - sqrt_term) * exp_term_2b
    )

    result = get_f2_longitudinal_velocity_correlation(
        collisional_radius, taylor_microscale, eulerian_integral_length
    )

    assert (
        result.shape == expected.shape
    )
    assert np.allclose(
        result, expected, atol=1e-10
    )


def test_invalid_inputs():
    """
    Ensure validation errors are raised for invalid inputs.
    """
    with pytest.raises(ValueError):
        get_f2_longitudinal_velocity_correlation(
            -0.1, 0.05, 1.0
        )  # Negative collisional_radius

    with pytest.raises(ValueError):
        get_f2_longitudinal_velocity_correlation(
            0.1, -0.05, 1.0
        )  # Negative taylor_microscale

    with pytest.raises(ValueError):
        get_f2_longitudinal_velocity_correlation(
            0.1, 0.05, -1.0
        )  # Negative eulerian_integral_length


def test_edge_cases():
    """
    Test compute_f2_longitudinal_velocity_correlation with extreme values.
    """
    collisional_radius = np.array(
        [1e-6, 1e-3, 10.0]
    )  # Very small and large values
    taylor_microscale = 0.05  # [m]
    eulerian_integral_length = 1.0  # [m]

    result = get_f2_longitudinal_velocity_correlation(
        collisional_radius, taylor_microscale, eulerian_integral_length
    )

    assert np.all(np.isfinite(result)), "Expected all values to be finite"



"""
Test module for the kinematic_viscosity module.
"""

import pytest
from particula.gas.properties.kinematic_viscosity import (
    get_kinematic_viscosity,
    get_kinematic_viscosity_via_system_state,
)


def test_kinematic_viscosity():
    """Test kinematic viscosity calculation."""
    dynamic_viscosity = 1.85e-5  # Pa*s
    fluid_density = 1.2  # kg/m^3
    expected_kinematic_viscosity = dynamic_viscosity / fluid_density
    assert (
        pytest.approx(
            get_kinematic_viscosity(dynamic_viscosity, fluid_density), 1e-5
        )
        == expected_kinematic_viscosity
    )


def test_kinematic_viscosity_via_system_state():
    """Test kinematic viscosity calculation via system state."""
    temperature = 300  # K
    fluid_density = 1.2  # kg/m^3
    expected_kinematic_viscosity = get_kinematic_viscosity_via_system_state(
        temperature, fluid_density
    )
    assert (
        pytest.approx(expected_kinematic_viscosity, 1e-5)
        == expected_kinematic_viscosity
    )


def test_kinematic_viscosity_zero_temperature():
    """Test for error handling with zero temperature."""
    with pytest.raises(ValueError):
        get_kinematic_viscosity_via_system_state(0, 1.2)


def test_kinematic_viscosity_negative_temperature():
    """Test for error handling with negative temperature."""
    with pytest.raises(ValueError):
        get_kinematic_viscosity_via_system_state(-10, 1.2)



"""
Sedimentation kernel for aerosol particles.

Seinfeld and Pandis (2016), Chapter 13, Equation 13A.4.
"""

from numpy.typing import NDArray
import numpy as np

from particula.particles.properties.settling_velocity import (
    particle_settling_velocity_via_system_state,
)
from particula.util.validate_inputs import validate_inputs


def calculate_collision_efficiency_function(
    radius1: float, radius2: float
) -> float:
    """
    Placeholder function to calculate collision efficiency.

    Args:
    -----
        - radius1 : Radius of the first particle [m].
        - radius2 : Radius of the second particle [m].

    Returns:
    --------
        - Collision efficiency [dimensionless].
    """
    # Implement the actual collision efficiency calculation here
    raise NotImplementedError(
        "Collision efficiency calculation not implemented yet."
    )


def sedimentation_sp2016(
    radius_particle: NDArray[np.float64],
    settling_velocities: NDArray[np.float64],
    calculate_collision_efficiency: bool = True,
) -> NDArray[np.float64]:
    """
    Calculate the sedimentation kernel for aerosol particles.

    Coagulation occurring due to gravitational settling when heavier particles
    settle faster than lighter ones, catching up and colliding with them. The
    coagulation coefficient is proportional to the product of the target area,
    the relative distance swept by the larger particle per unit time, and the
    collision efficiency

    Args:
    -----
        - radius_particle : Array of particle radii [m].
        - settling_velocities : Array of particle settling velocities [m/s].
        - calculate_collision_efficiency : Boolean to calculate collision
            efficiency or use 1.

    Returns:
    --------
        - Sedimentation kernel matrix for aerosol particles [m^3/s].

    References:
    - Seinfeld, J. H., & Pandis, S. N. (2016). Atmospheric chemistry and
        physics, Chapter 13, Equation 13A.4.
    """
    diameter_matrix = 2 * (
        radius_particle[:, np.newaxis] + radius_particle[np.newaxis, :]
    )
    velocity_diff_matrix = np.abs(
        settling_velocities[:, np.newaxis] - settling_velocities[np.newaxis, :]
    )

    if calculate_collision_efficiency:
        collision_efficiency_matrix = np.vectorize(
            calculate_collision_efficiency_function
        )(radius_particle[:, np.newaxis], radius_particle[np.newaxis, :])
    else:
        collision_efficiency_matrix = np.ones_like(diameter_matrix)

    return (
        np.pi
        / 4
        * diameter_matrix**2
        * velocity_diff_matrix
        * collision_efficiency_matrix
    )


@validate_inputs({"temperature": "positive", "pressure": "positive"})
def sedimentation_sp2016_via_system_state(
    radius_particle: NDArray[np.float64],
    density_particle: NDArray[np.float64],
    temperature: float,
    pressure: float,
    calculate_collision_efficiency: bool = True,
) -> NDArray[np.float64]:
    """
    Calculate the sedimentation kernel for aerosol particles via system state.

    Args:
    -----
        - radius_particle : Array of particle radii [m].
        - density_particle : Array of particle densities [kg/m³].
        - temperature : Temperature of the system [K].
        - pressure : Pressure of the system [Pa].
        - calculate_collision_efficiency : Boolean to calculate collision
            efficiency or use 1.

    Returns:
    --------
        - Sedimentation kernel matrix for aerosol particles [m^3/s].
    """
    settling_velocities = particle_settling_velocity_via_system_state(
        particle_radius=radius_particle,
        particle_density=density_particle,
        temperature=temperature,
        pressure=pressure,
    )

    return sedimentation_sp2016(
        radius_particle=radius_particle,
        settling_velocities=settling_velocities,
        calculate_collision_efficiency=calculate_collision_efficiency,
    )



"""Test Building ParticleRepresentation properties."""

import numpy as np
from particula.particles.representation import ParticleRepresentation
from particula.particles.distribution_strategies import (
    RadiiBasedMovingBin,
)
from particula.particles.surface_strategies import SurfaceStrategyVolume
from particula.particles.activity_strategies import ActivityIdealMass


# pylint: disable=too-many-arguments, too-many-positional-arguments
def setup_particle(
    strategy: RadiiBasedMovingBin = RadiiBasedMovingBin(),
    activity: ActivityIdealMass = ActivityIdealMass(),
    surface: SurfaceStrategyVolume = SurfaceStrategyVolume(),
    distribution: np.ndarray = np.array([1.0, 2.0, 3.0]),
    density: float = 1.0,
    concentration: np.ndarray = np.array([10, 20, 30]),
    charge: float = 1.0,
) -> ParticleRepresentation:
    """Setup ParticleRepresentation for testing with configurable parameters.

    - strategy : Strategy for particle distribution
    - activity : Activity strategy for particles
    - surface : Surface strategy for particles
    - distribution : Distribution array for particles
    - density : Density of particles
    - concentration : Concentration array for particles
    - charge : Charge of particles
    """
    return ParticleRepresentation(
        strategy=strategy,
        activity=activity,
        surface=surface,
        distribution=distribution,
        density=density,
        concentration=concentration,
        charge=charge,
    )


def test_get_mass():
    """Test get_mass method."""
    particle = setup_particle()
    mass = particle.get_mass()
    mass_clone = particle.get_mass(clone=True)

    assert isinstance(mass, np.ndarray)
    assert isinstance(mass_clone, np.ndarray)
    np.testing.assert_array_equal(mass, mass_clone)


def test_get_radius():
    """Test get_radius method."""
    particle = setup_particle()
    radius = particle.get_radius()
    radius_clone = particle.get_radius(clone=True)

    assert isinstance(radius, np.ndarray)
    assert isinstance(radius_clone, np.ndarray)
    np.testing.assert_array_equal(radius, radius_clone)


def test_get_mass_concentration():
    """Test get_mass_concentration method."""
    particle = setup_particle()
    mass = particle.get_mass()
    expected_mass_concentration = np.sum(mass*particle.get_concentration())
    total_mass = particle.get_mass_concentration()
    total_mass_clone = particle.get_mass_concentration(clone=True)

    assert isinstance(total_mass, np.float64)
    assert isinstance(total_mass_clone, np.float64)
    assert total_mass == total_mass_clone
    assert total_mass == expected_mass_concentration


def test_get_strategy():
    """Test get_strategy method."""
    particle = setup_particle()
    strategy = particle.get_strategy()
    strategy_clone = particle.get_strategy(clone=True)

    assert isinstance(strategy, RadiiBasedMovingBin)
    assert isinstance(strategy_clone, RadiiBasedMovingBin)


def test_get_activity():
    """Test get_activity method."""
    particle = setup_particle()
    activity = particle.get_activity()
    activity_clone = particle.get_activity(clone=True)

    assert isinstance(activity, ActivityIdealMass)
    assert isinstance(activity_clone, ActivityIdealMass)


def test_get_surface():
    """Test get_surface method."""
    particle = setup_particle()
    surface = particle.get_surface()
    surface_clone = particle.get_surface(clone=True)

    assert isinstance(surface, SurfaceStrategyVolume)
    assert isinstance(surface_clone, SurfaceStrategyVolume)


def test_get_distribution():
    """Test get_distribution method."""
    particle = setup_particle()
    distribution = particle.get_distribution()
    distribution_clone = particle.get_distribution(clone=True)

    np.testing.assert_array_equal(distribution, distribution_clone)


def test_get_density():
    """Test get_density method."""
    particle = setup_particle()
    density = particle.get_density()
    density_clone = particle.get_density(clone=True)

    np.testing.assert_array_equal(density, density_clone)


def test_get_concentration():
    """Test get_concentration method."""
    particle = setup_particle()
    concentration = particle.get_concentration()
    concentration_clone = particle.get_concentration(clone=True)

    np.testing.assert_array_equal(concentration, concentration_clone)


def test_get_charge():
    """Test get_charge method."""
    particle = setup_particle()
    charge = particle.get_charge()
    charge_clone = particle.get_charge(clone=True)

    np.testing.assert_array_equal(charge, charge_clone)


def test_get_volume():
    """Test get_volume method."""
    particle = setup_particle()
    volume = particle.get_volume()
    volume_clone = particle.get_volume(clone=True)

    assert volume == volume_clone


def test_get_species_mass():
    """Test get_species_mass method."""
    particle = setup_particle()
    species_mass = particle.get_species_mass()
    species_mass_clone = particle.get_species_mass(clone=True)

    np.testing.assert_array_equal(species_mass, species_mass_clone)


def test_get_total_concentration():
    """Test get_total_concentration method."""
    particle = setup_particle()
    conc = particle.get_concentration()
    expected_total_concentration = np.sum(conc)
    total_concentration = particle.get_total_concentration()
    total_concentration_clone = particle.get_total_concentration(clone=True)

    assert total_concentration == total_concentration_clone
    assert total_concentration == expected_total_concentration



"""Test the Kelvin effect module."""

import numpy as np
from particula.particles.properties.kelvin_effect_module import (
    kelvin_radius,
    kelvin_term,
)


def test_prop_kelvin_radius_scalar():
    """Test kelvin_radius function with scalar inputs."""
    effective_surface_tension = 0.072  # N/m
    effective_density = 1000  # kg/m^3
    molar_mass = 0.018  # kg/mol
    temperature = 298  # K
    expected_radius = (2 * effective_surface_tension * molar_mass) / (
        8.314 * temperature * effective_density
    )
    assert np.isclose(
        kelvin_radius(
            effective_surface_tension,
            effective_density,
            molar_mass,
            temperature,
        ),
        expected_radius,
    )


def test_prop_kelvin_radius_array():
    """Test kelvin_radius function with array inputs."""
    effective_surface_tension = np.array([0.072, 0.072])
    effective_density = np.array([1000, 1500])
    molar_mass = np.array([0.018, 0.018])
    temperature = 298
    expected_radius = (2 * effective_surface_tension * molar_mass) / (
        8.314 * temperature * effective_density
    )
    np.testing.assert_allclose(
        kelvin_radius(
            effective_surface_tension,
            effective_density,
            molar_mass,
            temperature,
        ),
        expected_radius,
        atol=1e-5,
    )


def test_prop_kelvin_term_scalar():
    """Test kelvin_term function with scalar inputs."""
    radius = 0.5  # m
    kelvin_radius_value = 0.1  # m
    expected_term = np.exp(kelvin_radius_value / radius)
    assert np.isclose(kelvin_term(radius, kelvin_radius_value), expected_term)


def test_prop_kelvin_term_broadcast():
    """Test kelvin_term function with broadcasting."""
    radius = np.array([0.5, 1.0, 5.0, 8.0])
    kelvin_radius_value = np.array([0.1, 0.2])
    expected_term = np.exp(
        kelvin_radius_value[np.newaxis, :] / radius[:, np.newaxis]
    )
    np.testing.assert_allclose(
        kelvin_term(radius, kelvin_radius_value), expected_term, rtol=1e-8
    )


def test_prop_kelvin_term_broadcast_single():
    """Test kelvin_term function with broadcasting."""
    radius = np.array([0.5, 1.0, 5.0, 8.0])
    kelvin_radius_value = np.array([0.1])
    expected_term = np.exp(kelvin_radius_value / radius)
    np.testing.assert_allclose(
        kelvin_term(radius, kelvin_radius_value), expected_term, rtol=1e-8
    )



# pylint: disable=duplicate-code
"""
Test the radial distribution of the coagulation kernel g12 using the
AO2008 DNS model.

By calling the get_g12_radial_distribution_ao2008 function, we also
test the private functions used in the calculation of the radial
distribution function.
"""

import pytest
import numpy as np

from particula.dynamics.coagulation.turbulent_dns_kernel.g12_radial_distribution_ao2008 import (
    get_g12_radial_distribution_ao2008,
)


def test_get_g12_radial_distribution_ao2008_scalar():
    """
    Test get_g12_radial_distribution_ao2008 with a small array input.
    """
    particle_radius = np.array(
        [10e-6, 20e-6, 30e-6]
    )  # Particle radii in meters
    stokes_number = np.array([0.5, 0.7, 1.2])  # Stokes numbers
    kolmogorov_length_scale = 1e-6  # m
    reynolds_lambda = 100  # Taylor-microscale Reynolds number
    normalized_accel_variance = 11  # Normalized acceleration variance
    kolmogorov_velocity = 0.01  # m/s
    kolmogorov_time = 0.005  # s

    # Compute g12
    g12_matrix = get_g12_radial_distribution_ao2008(
        particle_radius,
        stokes_number,
        kolmogorov_length_scale,
        reynolds_lambda,
        normalized_accel_variance,
        kolmogorov_velocity,
        kolmogorov_time,
    )

    # Check the shape
    assert g12_matrix.shape == (
        3,
        3,
    ), f"Expected shape (3,3), but got {g12_matrix.shape}"

    # Ensure values are non-negative
    assert np.all(g12_matrix >= 0), "Expected all values to be non-negative"


def test_get_g12_radial_distribution_ao2008_invalid_inputs():
    """
    Test that get_g12_radial_distribution_ao2008 raises validation errors
    for invalid inputs.
    """
    particle_radius = np.array([10e-6, 20e-6, 30e-6])
    stokes_number = np.array([0.5, 0.7, 1.2])
    kolmogorov_length_scale = 1e-6
    reynolds_lambda = 100
    normalized_accel_variance = 11
    kolmogorov_velocity = 0.01
    kolmogorov_time = 0.005

    with pytest.raises(ValueError):
        get_g12_radial_distribution_ao2008(
            -particle_radius,  # Negative values should raise an error
            stokes_number,
            kolmogorov_length_scale,
            reynolds_lambda,
            normalized_accel_variance,
            kolmogorov_velocity,
            kolmogorov_time,
        )

    with pytest.raises(ValueError):
        get_g12_radial_distribution_ao2008(
            particle_radius,
            -stokes_number,  # should raise an error
            kolmogorov_length_scale,
            reynolds_lambda,
            normalized_accel_variance,
            kolmogorov_velocity,
            kolmogorov_time,
        )

    with pytest.raises(ValueError):
        get_g12_radial_distribution_ao2008(
            particle_radius,
            stokes_number,
            -kolmogorov_length_scale,  # should raise an error
            reynolds_lambda,
            normalized_accel_variance,
            kolmogorov_velocity,
            kolmogorov_time,
        )

    with pytest.raises(ValueError):
        get_g12_radial_distribution_ao2008(
            particle_radius,
            stokes_number,
            kolmogorov_length_scale,
            -reynolds_lambda,  # should raise an error
            normalized_accel_variance,
            kolmogorov_velocity,
            kolmogorov_time,
        )


def test_get_g12_radial_distribution_ao2008_edge_cases():
    """
    Test get_g12_radial_distribution_ao2008 with edge cases, such as
    very small or very large Stokes numbers.
    """
    particle_radius = np.array([10e-6, 20e-6, 30e-6])
    stokes_number = np.array(
        [0.000001, 0.01, 2000.0]
    )  # Includes extreme Stokes numbers
    kolmogorov_length_scale = 1e-6
    reynolds_lambda = 100
    normalized_accel_variance = 11
    kolmogorov_velocity = 0.01
    kolmogorov_time = 0.005

    g12_matrix = get_g12_radial_distribution_ao2008(
        particle_radius,
        stokes_number,
        kolmogorov_length_scale,
        reynolds_lambda,
        normalized_accel_variance,
        kolmogorov_velocity,
        kolmogorov_time,
    )

    assert g12_matrix.shape == (
        3,
        3,
    ), f"Expected shape (3,3), but got {g12_matrix.shape}"
    assert np.all(g12_matrix >= 0), "Expected all values to be non-negative"



"""Tests for the diffusion coefficient property."""

import numpy as np
import pytest
from particula.particles.properties.diffusion_coefficient import (
    particle_diffusion_coefficient,
)


def test_particle_diffusion_coefficient_float():
    """Test the diffusion coefficient for a single float value."""
    temperature = 298.15
    particle_aerodynamic_mobility = 1e-9
    expected_diffusion_coefficient = 4.116404993e-30
    result = particle_diffusion_coefficient(
        temperature, particle_aerodynamic_mobility
    )
    assert result == pytest.approx(  # type: ignore
        expected_diffusion_coefficient
    )


def test_particle_diffusion_coefficient_array():
    """Test the diffusion coefficient for a numpy array."""
    temperature = np.array([298.15, 300.0, 310.0])
    particle_aerodynamic_mobility = np.array([1e-9, 2e-9, 3e-9])
    expected_diffusion_coefficient = np.array(
        [1.380649e-32, 2.761298e-32, 4.141947e-32]
    )
    result = particle_diffusion_coefficient(
        temperature, particle_aerodynamic_mobility
    )
    assert np.allclose(result, expected_diffusion_coefficient)


def test_particle_diffusion_coefficient_invalid():
    """Test the diffusion coefficient for invalid input."""
    temperature = "invalid"
    particle_aerodynamic_mobility = 1e-9
    with pytest.raises(TypeError):
        particle_diffusion_coefficient(
            temperature, particle_aerodynamic_mobility  # type: ignore
        )



"""Thermal Conductivity of air."""

import logging
from typing import Union
from numpy.typing import NDArray
import numpy as np

logger = logging.getLogger("particula")  # get instance of logger


def get_thermal_conductivity(
    temperature: Union[float, NDArray[np.float64]]
) -> Union[float, NDArray[np.float64]]:
    """
    Calculate the thermal conductivity of air as a function of temperature.
    Based on a simplified linear relation from atmospheric science literature.
    Only valid for temperatures within the range typically found on
    Earth's surface.

    Args:
    -----
    - temperature (Union[float, NDArray[np.float64]]): The temperature at which
    the thermal conductivity of air is to be calculated, in Kelvin (K).

    Returns:
    --------
    - Union[float, NDArray[np.float64]]: The thermal conductivity of air at the
    specified temperature in Watts per meter-Kelvin (W/m·K) or J/(m s K).

    Raises:
    ------
    - ValueError: If the temperature is below absolute zero (0 K).

    References:
    ----------
    - Seinfeld and Pandis, "Atmospheric Chemistry and Physics", Equation 17.54.
    """
    if np.any(temperature < 0):
        logger.error("Temperature must be greater than or equal to 0 Kelvin.")
        raise ValueError(
            "Temperature must be greater than or equal to 0 Kelvin."
        )
    return 1e-3 * (4.39 + 0.071 * temperature)



"""pytest tests for particle_resolved_method.py"""

import time
import numpy as np
from scipy.interpolate import RectBivariateSpline
from particula.dynamics.coagulation.particle_resolved_method import (
    interpolate_kernel,
    calculate_probabilities,
    resolve_final_coagulation_state,
    particle_resolved_update_step,
    particle_resolved_coagulation_step,
)


def test_interpolate_kernel():
    """Test the interpolate_kernel function."""
    kernel = np.random.rand(10, 10)
    kernel_radius = np.linspace(0, 1, 10)
    interp_func = interpolate_kernel(kernel, kernel_radius)
    assert isinstance(interp_func, RectBivariateSpline)


def test_calculate_probabilities():
    """Test the calculate_probabilities function."""
    kernel_values = np.array([1.0, 2.0], dtype=np.float64)
    time_step = 1.0
    events = 10
    tests = 5
    volume = 100.0
    probabilities = calculate_probabilities(
        kernel_values, time_step, events, tests, volume
    )
    expected_probabilities = (
        kernel_values * time_step * events / (tests * volume)
    )
    np.testing.assert_array_almost_equal(probabilities, expected_probabilities)


def test_resolve_final_coagulation_state():
    """Test the resolve_final_coagulation_state function."""
    small_indices = np.array([0, 1, 2], dtype=np.int64)
    large_indices = np.array([2, 3, 4], dtype=np.int64)
    particle_radius = np.array([1.0, 2.0, 3.0, 4.0, 5.0], dtype=np.float64)
    updated_small_indices, updated_large_indices = (
        resolve_final_coagulation_state(
            small_indices, large_indices, particle_radius
        )
    )
    assert len(updated_small_indices) == len(small_indices)
    assert len(updated_large_indices) == len(large_indices)
    assert np.all(updated_large_indices >= updated_small_indices)


def test_resolve_final_coagulation_state_large():
    """Test the resolve_final_coagulation_state function with a
    large number of particles."""
    num_particles = 100_000
    small_indices = np.random.randint(
        0, num_particles, size=num_particles // 2, dtype=np.int64
    )
    large_indices = np.random.randint(
        0, num_particles, size=num_particles // 2, dtype=np.int64
    )
    particle_radius = np.random.uniform(1.0, 10.0, size=num_particles).astype(
        np.float64
    )

    # Introduce duplicate collisions
    small_indices[:10] = large_indices[:10]

    # Time the function
    start_time = time.time()
    updated_small_indices, updated_large_indices = (
        resolve_final_coagulation_state(
            small_indices, large_indices, particle_radius
        )
    )
    end_time = time.time()
    elapsed_time = end_time - start_time
    time_per_particle = elapsed_time / num_particles
    current_version_time_per_particle = 1e-05

    # incase we change the implementation and it is slower
    assert time_per_particle <= current_version_time_per_particle

    assert len(updated_small_indices) == len(small_indices)
    assert len(updated_large_indices) == len(large_indices)


def test_particle_resolved_update_step():
    """Test the particle_resolved_update_step function."""
    particle_radius = np.array([1.0, 2.0, 3.0], dtype=np.float64)
    loss = np.zeros_like(particle_radius)
    gain = np.zeros_like(particle_radius)
    small_index = np.array([0], dtype=np.int64)
    large_index = np.array([1], dtype=np.int64)
    updated_radius, updated_loss, updated_gain = particle_resolved_update_step(
        particle_radius, loss, gain, small_index, large_index
    )
    assert updated_radius[0] == 0
    assert updated_radius[1] > 2.0
    assert updated_loss[0] == 1.0
    assert updated_gain[1] == 2.0


def test_particle_resolved_coagulation_step():
    """Test the particle_resolved_coagulation_step function."""
    particle_radius = np.array([1.0, 2.0, 3.0], dtype=np.float64)
    kernel = np.random.rand(10, 10)
    kernel_radius = np.linspace(0, 1, 10)
    volume = 100.0
    time_step = 1.0
    random_generator = np.random.default_rng(seed=42)
    loss_gain_index = particle_resolved_coagulation_step(
        particle_radius,
        kernel,
        kernel_radius,
        volume,
        time_step,
        random_generator,
    )
    assert loss_gain_index.shape[1] == 2
    assert np.all(loss_gain_index[:, 0] < loss_gain_index[:, 1])


def test_particle_resolved_coagulation_step_empty_array():
    """Test the particle_resolved_coagulation_step function with an empty array."""
    particle_radius = np.array([], dtype=np.float64)
    kernel = np.random.rand(10, 10)
    kernel_radius = np.linspace(0, 1, 10)
    volume = 100.0
    time_step = 1.0
    random_generator = np.random.default_rng(seed=42)
    loss_gain_index = particle_resolved_coagulation_step(
        particle_radius,
        kernel,
        kernel_radius,
        volume,
        time_step,
        random_generator,
    )
    assert loss_gain_index.size == 0


def test_particle_resolved_coagulation_step_duplicate_collisions():
    """Test the particle_resolved_coagulation_step function with all duplicate collisions."""
    particle_radius = np.array([1.0, 1.0, 1.0], dtype=np.float64)
    kernel = np.random.rand(10, 10)
    kernel_radius = np.linspace(0, 1, 10)
    volume = 100.0
    time_step = 1.0
    random_generator = np.random.default_rng(seed=42)
    loss_gain_index = particle_resolved_coagulation_step(
        particle_radius,
        kernel,
        kernel_radius,
        volume,
        time_step,
        random_generator,
    )
    assert loss_gain_index.shape[1] == 2
    assert np.all(loss_gain_index[:, 0] < loss_gain_index[:, 1])
    assert np.all(particle_radius[loss_gain_index[:, 0]] == 0)
    assert np.all(particle_radius[loss_gain_index[:, 1]] > 1.0)



""" This module contains tests for the validate_inputs decorator.
"""

import pytest

from particula.util.validate_inputs import validate_inputs

# Define a sample function to test the decorator
some_dict = {
    "x": "positive",
    "y": "negative",
    "z": "nonpositive",
    "w": "nonnegative",
    "v": "nonzero",
}


@validate_inputs(some_dict)
def sample_function(x, y, z, w, v):
    """ A sample function to test the decorator """
    return x, y, z, w, v


def test_validate_positive():
    """ Test that the decorator raises an error when a positive argument is negative """
    with pytest.raises(ValueError, match="Argument 'x' must be positive."):
        sample_function(x=-1, y=-1, z=0, w=0, v=1)


def test_validate_negative():
    """ Test that the decorator raises an error when a negative argument is positive """
    with pytest.raises(ValueError, match="Argument 'y' must be negative."):
        sample_function(1, 1, 0, 0, 1)


def test_validate_nonpositive():
    """ Test that the decorator raises an error when a nonpositive argument is positive """
    with pytest.raises(ValueError, match="Argument 'z' must be nonpositive."):
        sample_function(1, -1, 1, 0, 1)


def test_validate_nonnegative():
    """ Test that the decorator raises an error when a nonnegative argument is negative """
    with pytest.raises(ValueError, match="Argument 'w' must be nonnegative."):
        sample_function(1, -1, 0, -1, 1)


def test_validate_nonzero():
    """ Test that the decorator raises an error when a nonzero argument is zero """
    with pytest.raises(ValueError, match="Argument 'v' must be nonzero."):
        sample_function(1, -1, 0, 0, 0)


def test_valid_inputs():
    """" Test that the decorator does not raise an error when all arguments are valid """
    assert sample_function(1, -1, 0, 0, 1) == (1, -1, 0, 0, 1)



"""Coagulation functions exposed via __init__.

If you want specific sub functions, import them directly from
`particula.next.dynamics.coagulation.xxx.xxx` instead.
"""

# pylint: disable=unused-import, disable=line-too-long
# flake8: noqa
# pyright: basic

from particula.dynamics.coagulation.brownian_kernel import (
    brownian_coagulation_kernel,
    brownian_coagulation_kernel_via_system_state,
)
from particula.dynamics.coagulation.kernel import (
    HardSphere,
    CoulombDyachkov2007,
    CoulombGatti2008,
    CoulombGopalakrishnan2012,
    CoulumbChahl2019,
)
from particula.dynamics.coagulation.particle_resolved_method import (
    particle_resolved_coagulation_step,
)
from particula.dynamics.coagulation.super_droplet_method import (
    super_droplet_coagulation_step,
)
from particula.dynamics.coagulation.strategy import (
    DiscreteSimple,
    DiscreteGeneral,
    ContinuousGeneralPDF,
    ParticleResolved,
)
from particula.dynamics.coagulation.rate import (
    continuous_loss,
    discrete_loss,
    discrete_gain,
    continuous_gain,
)
from particula.dynamics.coagulation.transition_regime import (
    hard_sphere,
    coulomb_dyachkov2007,
    coulomb_gatti2008,
    coulomb_gopalakrishnan2012,
    coulomb_chahl2019,
)



""" Module for calculating Knudsen number
"""

from typing import Union
import logging
from numpy.typing import NDArray
import numpy as np

logger = logging.getLogger("particula")


def calculate_knudsen_number(
    mean_free_path: Union[float, NDArray[np.float64]],
    particle_radius: Union[float, NDArray[np.float64]],
) -> Union[float, NDArray[np.float64]]:
    """
    Calculate the Knudsen number using the mean free path of the gas and the
    radius of the particle. The Knudsen number is a dimensionless number that
    indicates the regime of gas flow relative to the size of particles.

    Args:
    -----
    - mean_free_path (Union[float, NDArray[np.float64]]): The mean free path of
    the gas molecules [meters (m)].
    - particle_radius (Union[float, NDArray[np.float64]]): The radius of the
    particle [meters (m)].

    Returns:
    --------
    - Union[float, NDArray[np.float64]]: The Knudsen number, which is the
    ratio of the mean free path to the particle radius.

    References:
    -----------
    - For more information at https://en.wikipedia.org/wiki/Knudsen_number
    """
    if not isinstance(mean_free_path, (float, np.ndarray)) or not isinstance(
        particle_radius, (float, np.ndarray)
    ):
        message = "The input must be a float or a numpy array"
        logger.error(message)
        raise TypeError(message)
    if isinstance(mean_free_path, float) and isinstance(
        particle_radius, (float, np.ndarray)
    ):
        return mean_free_path / particle_radius

    if isinstance(mean_free_path, np.ndarray) and isinstance(
        particle_radius, np.ndarray
    ):
        if (
            (mean_free_path.size == particle_radius.size)
            or (mean_free_path.size == 1)
            or (particle_radius.size == 1)
        ):
            return mean_free_path / particle_radius

        # Reshape to (n, 1) and vector_b to (1, m) to broadcast (n, m)
        particle_radius = particle_radius[
            :, np.newaxis
        ]  # Adds a new axis, creating a column vector
        mean_free_path = mean_free_path[
            np.newaxis, :
        ]  # Adds a new axis, creating a row vector
        return mean_free_path / particle_radius

    message = (
        "The input arrays must have the same size"
        " or one of them must have size 1"
    )
    logger.error(message)
    raise ValueError(message)



"""Tests for the size_distribution_convert module."""
import numpy as np


from particula.util.size_distribution_convert import (
    SizerConverter,
    get_conversion_strategy,
    DNdlogDPtoPMSConversionStrategy,
    PMStoPDFConversionStrategy,
    DNdlogDPtoPDFConversionStrategy
)


def test_dndlogdp_to_pms_conversion_strategy():
    """Test the conversion from dn/dlogdp to PMS format."""
    strategy = DNdlogDPtoPMSConversionStrategy()
    diameters = np.array([1, 2, 3])
    concentration = np.array([0.1, 0.2, 0.3])
    expected = np.array([0.04771213, 0.04436975, 0.04383841])
    # call function
    converted_concentration = strategy.convert(diameters, concentration)
    assert np.allclose(converted_concentration, expected, atol=1e-6)
    # call inverse function
    converted_inverse = strategy.convert(
        diameters, converted_concentration, inverse=True)
    assert np.allclose(converted_inverse, concentration, atol=1e-6)


def test_pms_to_pdf_conversion_strategy():
    """Test the conversion from PMS to PDF format."""
    strategy = PMStoPDFConversionStrategy()
    diameters = np.array([1, 10, 30])
    concentration = np.array([0.1, 0.2, 0.3])
    expected = np.array([0.01111111, 0.01, 0.00681818])
    # call function
    converted_concentration = strategy.convert(diameters, concentration)
    assert np.allclose(converted_concentration, expected, atol=1e-6)
    # call inverse function
    converted_inverse = strategy.convert(
        diameters, converted_concentration, inverse=True)
    assert np.allclose(converted_inverse, concentration, atol=1e-6)


def test_dndlogdp_to_pdf_conversion_strategy():
    """Test the conversion from dn/dlogdp to PDF format."""
    strategy = DNdlogDPtoPDFConversionStrategy()
    diameters = np.array([1, 2, 3])
    concentration = np.array([0.1, 0.2, 0.3])
    expected = np.array([0.04771213, 0.04436975, 0.04383841])
    # call function
    converted_concentration = strategy.convert(diameters, concentration)
    assert np.allclose(converted_concentration, expected, atol=1e-6)
    # call inverse function
    converted_inverse = strategy.convert(
        diameters, converted_concentration, inverse=True)
    assert np.allclose(converted_inverse, concentration, atol=1e-6)


def test_converter():
    """Test the Converter class."""
    strategy = DNdlogDPtoPMSConversionStrategy()
    converter = SizerConverter(strategy)
    diameters = np.array([1, 2, 3])
    concentration = np.array([0.1, 0.2, 0.3])
    expected = np.array([0.04771213, 0.04436975, 0.04383841])
    # call function
    converted_concentration = converter.convert(diameters, concentration)
    assert np.allclose(converted_concentration, expected, atol=1e-6)
    # call inverse function
    converted_inverse = converter.convert(
        diameters, converted_concentration, inverse=True)
    assert np.allclose(converted_inverse, concentration, atol=1e-6)


def test_get_conversion_strategy():
    """Test the get_conversion_strategy function."""
    strategy = get_conversion_strategy('dn/dlogdp', 'pms')
    assert isinstance(strategy, DNdlogDPtoPMSConversionStrategy)

    strategy = get_conversion_strategy('pms', 'pdf')
    assert isinstance(strategy, PMStoPDFConversionStrategy)

    strategy = get_conversion_strategy('dn/dlogdp', 'pdf')
    assert isinstance(strategy, DNdlogDPtoPDFConversionStrategy)

    try:
        strategy = get_conversion_strategy('invalid', 'pms')
        assert False  # Should raise a ValueError
    except ValueError:
        assert True

    try:
        strategy = get_conversion_strategy('dn/dlogdp', 'invalid')
        assert False  # Should raise a ValueError
    except ValueError:
        assert True



""" validate inputs
"""

import inspect
from functools import wraps

import numpy as np


def validate_positive(value, name):
    """ validate positive """
    if np.any(value <= 0):
        raise ValueError(f"Argument '{name}' must be positive.")


def validate_negative(value, name):
    """ validate negative """
    if np.any(value >= 0):
        raise ValueError(f"Argument '{name}' must be negative.")


def validate_nonpositive(value, name):
    """ validate nonpositive """
    if np.any(value > 0):
        raise ValueError(f"Argument '{name}' must be nonpositive.")


def validate_nonnegative(value, name):
    """ validate nonnegative """
    if np.any(value < 0):
        raise ValueError(f"Argument '{name}' must be nonnegative.")


def validate_nonzero(value, name):
    """ validate nonzero """
    if np.any(value == 0):
        raise ValueError(f"Argument '{name}' must be nonzero.")


def validate_inputs(dict_args):
    """
    A decorator to validate that specified arguments meet certain constraints.

    Arguments:
        - dict_args : Dictionary of argument names and their constraints.
            {"arg1": "positive", "arg2": "negative", ...}
            - Options: "positive", "negative", "nonpositive", "nonnegative",
                "nonzero".


    Returns:
        A decorator for input validation.
    """
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            # Map argument names to values
            sig = inspect.signature(func)
            params = list(sig.parameters.keys())
            for name, comp in dict_args.items():
                value = kwargs.get(
                    name, args[params.index(name)]
                    if name in params and params.index(name) < len(args)
                    else None
                )
                if comp == "positive":
                    validate_positive(value, name)
                elif comp == "negative":
                    validate_negative(value, name)
                elif comp == "nonpositive":
                    validate_nonpositive(value, name)
                elif comp == "nonnegative":
                    validate_nonnegative(value, name)
                elif comp == "nonzero":
                    validate_nonzero(value, name)
                else:
                    raise ValueError(
                        f"Unknown validation '{comp}' for argument '{name}'."
                    )
            return func(*args, **kwargs)
        return wrapper
    return decorator



"""Tests for the aerodynamic size module."""

import numpy as np
import pytest
from particula.particles.properties.aerodynamic_size import (
    particle_aerodynamic_length,
    get_aerodynamic_shape_factor,
)


def test_particle_aerodynamic_length_single_value():
    """
    Verify that the particle_aerodynamic_length function calculates the
    correct aerodynamic length for a single particle.
    """
    physical_length = 0.00005  # 50 micrometers
    physical_slip_correction_factor = 1.1
    aerodynamic_slip_correction_factor = 1.0
    density = 1000  # kg/m^3
    expected_length = physical_length * np.sqrt(
        (physical_slip_correction_factor / aerodynamic_slip_correction_factor)
        * (density / 1000)
    )
    actual_length = particle_aerodynamic_length(
        physical_length,
        physical_slip_correction_factor,
        aerodynamic_slip_correction_factor,
        density,
    )
    assert np.isclose(
        actual_length, expected_length
    ), "The value does not match."


def test_particle_aerodynamic_length_array_input():
    """
    Test that the particle_aerodynamic_length function handles numpy array
    inputs correctly.
    """
    physical_length = np.array([0.00005, 0.00007])  # Array of lengths
    physical_slip_correction_factor = np.array([1.1, 1.2])
    aerodynamic_slip_correction_factor = np.array([1.0, 1.1])
    density = np.array([1000, 1200])
    expected_length = physical_length * np.sqrt(
        (physical_slip_correction_factor / aerodynamic_slip_correction_factor)
        * (density / 1000)
    )
    actual_length = particle_aerodynamic_length(
        physical_length,
        physical_slip_correction_factor,
        aerodynamic_slip_correction_factor,
        density,
    )
    assert np.allclose(actual_length, expected_length)


@pytest.mark.parametrize("shape_key, expected_factor", [
    ("sphere", 1.0),
    ("cube", 1.08),
    ("cylinder_avg_aspect_2", 1.1),
    ("cylinder_avg_aspect_5", 1.35),
    ("cylinder_avg_aspect_10", 1.68),
    ("spheres_cluster_3", 1.15),
    ("spheres_cluster_4", 1.17),
    ("bituminous_coal", 1.08),
    ("quartz", 1.36),
    ("sand", 1.57),
    ("talc", 1.88)
])
def test_get_aerodynamic_shape_factor(shape_key, expected_factor):
    """Test the get_aerodynamic_shape_factor function."""
    actual_factor = get_aerodynamic_shape_factor(shape_key)
    assert actual_factor == expected_factor, "The shape factor does not match."


def test_get_aerodynamic_shape_factor_invalid():
    """
    Test that the get_aerodynamic_shape_factor function raises a ValueError
    for unknown shapes.
    """
    with pytest.raises(ValueError):
        get_aerodynamic_shape_factor("unknown_shape")



"""
Tests that the collision kernel function can be evaluated get_kernel_ao2008.
"""

import unittest
import numpy as np
from particula.dynamics.coagulation.turbulent_dns_kernel.kernel_ao2008 import (
    get_kernel_ao2008,
)


class TestKernelAO2008(unittest.TestCase):
    # pylint: disable=too-many-instance-attributes
    """Unit tests for the collision kernel function get_kernel_ao2008."""

    def setUp(self):
        """Set up common test parameters."""
        self.stokes_number_scalar = 0.5  # [-]
        self.kolmogorov_length_scale = 1e-6  # [m]
        self.reynolds_lambda = 100  # [-]
        self.normalized_accel_variance = 11  # [-]
        self.kolmogorov_velocity = 0.01  # [m/s]
        self.kolmogorov_time = 0.005  # [s]

        # Array-based test inputs
        self.particle_radius_array = np.array([10e-6, 20e-6, 30e-6])  # [m]
        self.particle_inertia_time_array = np.array([0.02, 0.03, 0.05])  # [s]
        self.stokes_number_array = np.array([0.5, 0.6, 0.7])  # [-]
        self.velocity_dispersion_scalar = 0.1  # [m/s]

    def test_get_kernel_ao2008_array(self):
        """Test get_kernel_ao2008 with NumPy array inputs."""
        result = get_kernel_ao2008(
            self.particle_radius_array,
            self.velocity_dispersion_scalar,
            self.particle_inertia_time_array,
            self.stokes_number_array,
            self.kolmogorov_length_scale,
            self.reynolds_lambda,
            self.normalized_accel_variance,
            self.kolmogorov_velocity,
            self.kolmogorov_time,
        )

        self.assertEqual(
            result.shape,
            (3, 3),
            f"Expected shape (3, 3), but got {result.shape}",
        )

    def test_invalid_inputs(self):
        """Ensure validation errors are raised for invalid inputs."""
        with self.assertRaises(ValueError):
            get_kernel_ao2008(
                -1*self.particle_radius_array,
                self.velocity_dispersion_scalar,
                self.particle_inertia_time_array,
                self.stokes_number_array,
                self.kolmogorov_length_scale,
                self.reynolds_lambda,
                self.normalized_accel_variance,
                self.kolmogorov_velocity,
                self.kolmogorov_time,
            )  # Negative radius

        with self.assertRaises(ValueError):
            get_kernel_ao2008(
                self.particle_radius_array,
                -1*self.velocity_dispersion_scalar,
                self.particle_inertia_time_array,
                self.stokes_number_array,
                self.kolmogorov_length_scale,
                self.reynolds_lambda,
                self.normalized_accel_variance,
                self.kolmogorov_velocity,
                self.kolmogorov_time,
            )  # Negative velocity_dispersion

        with self.assertRaises(ValueError):
            get_kernel_ao2008(
                self.particle_radius_array,
                self.velocity_dispersion_scalar,
                -1*self.particle_inertia_time_array,
                self.stokes_number_array,
                self.kolmogorov_length_scale,
                self.reynolds_lambda,
                self.normalized_accel_variance,
                self.kolmogorov_velocity,
                self.kolmogorov_time,
            )  # Negative particle_inertia_time

        with self.assertRaises(ValueError):
            get_kernel_ao2008(
                self.particle_radius_array,
                self.velocity_dispersion_scalar,
                self.particle_inertia_time_array,
                -1*self.stokes_number_array,
                self.kolmogorov_length_scale,
                self.reynolds_lambda,
                self.normalized_accel_variance,
                self.kolmogorov_velocity,
                self.kolmogorov_time,
            )  # Negative stokes_number



""""Surface tension strategies factory.
"""

from typing import Union
from particula.abc_factory import StrategyFactory
from particula.particles.surface_builders import (
    SurfaceStrategyMolarBuilder,
    SurfaceStrategyMassBuilder,
    SurfaceStrategyVolumeBuilder,
)
from particula.particles.surface_strategies import (
    SurfaceStrategyVolume,
    SurfaceStrategyMass,
    SurfaceStrategyMolar,
)


class SurfaceFactory(
    StrategyFactory[
        Union[
            SurfaceStrategyVolumeBuilder,
            SurfaceStrategyMassBuilder,
            SurfaceStrategyMolarBuilder,
        ],
        Union[
            SurfaceStrategyVolume, SurfaceStrategyMass, SurfaceStrategyMolar
        ],
    ]
):
    """Factory class to call and create surface tension strategies.

    Factory class to create surface tension strategy builders for
    calculating surface tension and the Kelvin effect for species in
    particulate phases.

    Methods:
        get_builders(): Returns the mapping of strategy types to builder
        instances.
        get_strategy(strategy_type, parameters): Gets the strategy instance
        for the specified strategy type.
            strategy_type: Type of surface tension strategy to use, can be
            'volume', 'mass', or 'molar'.
            parameters(Dict[str, Any], optional): Parameters required for the
            builder, dependent on the chosen strategy type.
                volume: density, surface_tension
                mass: density, surface_tension
                molar: molar_mass, density, surface_tension

    Returns:
        SurfaceStrategy: An instance of the specified SurfaceStrategy.

    Raises:
        ValueError: If an unknown strategy type is provided.
        ValueError: If any required key is missing during check_keys or
            pre_build_check, or if trying to set an invalid parameter.
    """

    def get_builders(self):
        """
        Returns the mapping of strategy types to builder instances.

        Returns:
            Dict[str, BuilderT]: A dictionary mapping strategy types to
            builder instances.
                volume: SurfaceStrategyVolumeBuilder
                mass: SurfaceStrategyMassBuilder
                molar: SurfaceStrategyMolarBuilder
        """
        return {
            "volume": SurfaceStrategyVolumeBuilder(),
            "mass": SurfaceStrategyMassBuilder(),
            "molar": SurfaceStrategyMolarBuilder(),
        }



"""
Velocity correlation terms for the two-point velocity correlation function
f₂(R) from Ayala et al. (2008).
"""

from typing import Union
import numpy as np
from numpy.typing import NDArray

from particula.util.validate_inputs import validate_inputs
from particula.dynamics.coagulation.turbulent_dns_kernel.velocity_correlation_terms_ao2008 import (
    compute_beta,
)


@validate_inputs(
    {
        "collisional_radius": "positive",
        "taylor_microscale": "positive",
        "eulerian_integral_length": "positive",
    }
)
def get_f2_longitudinal_velocity_correlation(
    collisional_radius: Union[float, NDArray[np.float64]],
    taylor_microscale: Union[float, NDArray[np.float64]],
    eulerian_integral_length: Union[float, NDArray[np.float64]],
) -> Union[float, NDArray[np.float64]]:
    """
    Compute the longitudinal two-point velocity correlation function f₂(R).

    This function describes the correlation of velocity fluctuations as a
    function of collisional radius R between two colliding droplets.

    The function is given by:

        f₂(R) =  1 / (2 sqrt(1 - 2β²))\
                 × { (1 + sqrt(1 - 2β²)) exp[-2R / ((1 + sqrt(1 - 2β²)) L_e)]\
                 - (1 - sqrt(1 - 2β²)) exp[-2R / ((1 - sqrt(1 - 2β²)) L_e)] }

    - β = (sqrt(2) * λ) / L_e
    - λ (taylor_microscale) : Taylor microscale [m].
    - L_e (eulerian_integral_length) : Eulerian integral length scale [m].
    - R (collisional_radius) : Distance between colliding droplets [m].

    Arguments:
    ----------
        - collisional_radius : Distance between two colliding droplets [m].
        - taylor_microscale : Taylor microscale [m].
        - eulerian_integral_length : Eulerian integral length scale [m].

    Returns:
    --------
        - f₂(R) value [-].

    References:
    -----------
    - Ayala, O., Rosa, B., & Wang, L. P. (2008). Effects of turbulence on
        the geometric collision rate of sedimenting droplets. Part 2.
        Theory and parameterization. New Journal of Physics, 10.
        https://doi.org/10.1088/1367-2630/10/7/075016
    """
    beta = compute_beta(taylor_microscale, eulerian_integral_length)

    sqrt_term = np.sqrt(1 - 2 * beta**2)
    denominator = 2 * sqrt_term

    # Compute exponential terms
    exp_term_1 = np.exp(
        -2 * collisional_radius / ((1 + sqrt_term) * eulerian_integral_length)
    )
    exp_term_2 = np.exp(
        -2 * collisional_radius / ((1 - sqrt_term) * eulerian_integral_length)
    )

    # Compute f₂(R)
    return (1 / denominator) * (
        (1 + sqrt_term) * exp_term_1 - (1 - sqrt_term) * exp_term_2
    )



"""
Test module for the Taylor microscale functions.
"""

import unittest

import pytest
import numpy as np

from particula.gas.properties.taylor_microscale_module import (
    get_lagrangian_taylor_microscale_time,
    get_taylor_microscale,
    get_taylor_microscale_reynolds_number,
)


def test_get_lagrangian_taylor_microscale_time_scalar():
    """
    Test get_lagrangian_taylor_microscale_time with scalar inputs.
    """
    kolmogorov_time = 0.01  # s
    re_lambda = 100  # Reynolds number
    ao = 11  # Normalized acceleration variance

    expected = kolmogorov_time * np.sqrt((2 * re_lambda) / (15**0.5 * ao))
    result = get_lagrangian_taylor_microscale_time(
        kolmogorov_time, re_lambda, ao
    )

    assert np.isclose(result, expected, atol=1e-10)


def test_get_lagrangian_taylor_microscale_time_array():
    """
    Test get_lagrangian_taylor_microscale_time with NumPy array inputs.
    """
    kolmogorov_time = np.array([0.01, 0.02])
    re_lambda = np.array([100, 200])
    ao = np.array([11, 12])

    expected = kolmogorov_time * np.sqrt((2 * re_lambda) / (15**0.5 * ao))
    result = get_lagrangian_taylor_microscale_time(
        kolmogorov_time, re_lambda, ao
    )

    assert np.allclose(result, expected, atol=1e-10)


def test_get_lagrangian_taylor_microscale_time_invalid():
    """
    Test that get_lagrangian_taylor_microscale_time raises errors for invalid
    inputs.
    """
    with pytest.raises(ValueError):
        get_lagrangian_taylor_microscale_time(-0.01, 100, 11)

    with pytest.raises(ValueError):
        get_lagrangian_taylor_microscale_time(0.01, -100, 11)

    with pytest.raises(ValueError):
        get_lagrangian_taylor_microscale_time(0.01, 100, -11)


def test_get_taylor_microscale_scalar():
    """
    Test get_taylor_microscale with scalar inputs.
    """
    rms_velocity = 0.5  # m/s
    kinematic_viscosity = 1.5e-5  # m²/s
    turbulent_dissipation = 1e-3  # m²/s³

    expected = rms_velocity * np.sqrt(
        (15 * kinematic_viscosity**2) / turbulent_dissipation
    )
    result = get_taylor_microscale(
        rms_velocity, kinematic_viscosity, turbulent_dissipation
    )

    assert np.isclose(result, expected, atol=1e-10)


def test_get_taylor_microscale_array():
    """
    Test get_taylor_microscale with NumPy array inputs.
    """
    rms_velocity = np.array([0.5, 0.8])
    kinematic_viscosity = np.array([1.5e-5, 1.2e-5])
    turbulent_dissipation = np.array([1e-3, 2e-3])

    expected = rms_velocity * np.sqrt(
        (15 * kinematic_viscosity**2) / turbulent_dissipation
    )
    result = get_taylor_microscale(
        rms_velocity, kinematic_viscosity, turbulent_dissipation
    )

    assert np.allclose(result, expected, atol=1e-10)


def test_get_taylor_microscale_invalid():
    """
    Test that get_taylor_microscale raises errors for invalid inputs.
    """
    with pytest.raises(ValueError):
        get_taylor_microscale(-0.5, 1.5e-5, 1e-3)

    with pytest.raises(ValueError):
        get_taylor_microscale(0.5, -1.5e-5, 1e-3)

    with pytest.raises(ValueError):
        get_taylor_microscale(0.5, 1.5e-5, -1e-3)


def test_taylor_scales_edge_case():
    """
    Test both Taylor scale functions with very small values near machine
    precision.
    """
    kolmogorov_time = 1e-10
    re_lambda = 1.0
    ao = 1.0

    expected_tt = kolmogorov_time * np.sqrt((2 * re_lambda) / (15**0.5 * ao))
    result_tt = get_lagrangian_taylor_microscale_time(
        kolmogorov_time, re_lambda, ao
    )

    assert np.isclose(result_tt, expected_tt, atol=1e-10)

    rms_velocity = 1e-10
    kinematic_viscosity = 1e-10
    turbulent_dissipation = 1e-10

    expected_lambda = rms_velocity * np.sqrt(
        (15 * kinematic_viscosity**2) / turbulent_dissipation
    )
    result_lambda = get_taylor_microscale(
        rms_velocity, kinematic_viscosity, turbulent_dissipation
    )

    assert np.isclose(result_lambda, expected_lambda, atol=1e-10)


class TestTaylorMicroscaleReynolds(unittest.TestCase):
    """Unit tests for the Taylor-microscale Reynolds number function."""

    def setUp(self):
        """Set up common test parameters."""
        self.fluid_rms_velocity_scalar = 0.5  # [m/s]
        self.taylor_microscale_scalar = 0.05  # [m]
        self.kinematic_viscosity_scalar = 1.5e-5  # [m²/s]

        self.fluid_rms_velocity_array = np.array([0.5, 1.0, 1.5])  # [m/s]
        self.taylor_microscale_array = np.array([0.05, 0.08, 0.1])  # [m]
        self.kinematic_viscosity_array = np.array(
            [1.5e-5, 1.2e-5, 1.0e-5]
        )  # [m²/s]

    def test_get_taylor_microscale_reynolds_number_scalar(self):
        """Test with scalar inputs."""
        expected = (
            self.fluid_rms_velocity_scalar
            * self.taylor_microscale_scalar
            / self.kinematic_viscosity_scalar
        )

        result = get_taylor_microscale_reynolds_number(
            self.fluid_rms_velocity_scalar,
            self.taylor_microscale_scalar,
            self.kinematic_viscosity_scalar,
        )

        self.assertAlmostEqual(result, expected, places=10)

    def test_get_taylor_microscale_reynolds_number_array(self):
        """Test with NumPy array inputs."""
        expected = (
            self.fluid_rms_velocity_array
            * self.taylor_microscale_array
            / self.kinematic_viscosity_array
        )

        result = get_taylor_microscale_reynolds_number(
            self.fluid_rms_velocity_array,
            self.taylor_microscale_array,
            self.kinematic_viscosity_array,
        )

        np.testing.assert_allclose(result, expected, atol=1e-10)

    def test_invalid_inputs(self):
        """Ensure validation errors are raised for invalid inputs."""
        with self.assertRaises(ValueError):
            get_taylor_microscale_reynolds_number(
                -0.5, 0.05, 1.5e-5
            )  # Negative fluid RMS velocity

        with self.assertRaises(ValueError):
            get_taylor_microscale_reynolds_number(
                0.5, -0.05, 1.5e-5
            )  # Negative taylor microscale

        with self.assertRaises(ValueError):
            get_taylor_microscale_reynolds_number(
                0.5, 0.05, -1.5e-5
            )  # Negative kinematic viscosity

    def test_edge_cases(self):
        """Test with extreme values (very small and very large)."""
        fluid_rms_velocity = np.array([1e-6, 1e-3, 10.0])  # [m/s]
        taylor_microscale = np.array([1e-6, 1e-3, 10.0])  # [m]
        kinematic_viscosity = np.array([1e-6, 1e-3, 10.0])  # [m²/s]

        result = get_taylor_microscale_reynolds_number(
            fluid_rms_velocity, taylor_microscale, kinematic_viscosity
        )

        self.assertTrue(np.all(np.isfinite(result)))

    def test_consistency(self):
        """Ensure function produces consistent results for repeated inputs."""
        result1 = get_taylor_microscale_reynolds_number(
            self.fluid_rms_velocity_array,
            self.taylor_microscale_array,
            self.kinematic_viscosity_array,
        )

        result2 = get_taylor_microscale_reynolds_number(
            self.fluid_rms_velocity_array,
            self.taylor_microscale_array,
            self.kinematic_viscosity_array,
        )

        np.testing.assert_allclose(result1, result2, atol=1e-10)



"""
Stokes number calculation.
"""

from typing import Union
from numpy.typing import NDArray
import numpy as np

from particula.util.validate_inputs import validate_inputs


@validate_inputs(
    {
        "particle_inertia_time": "positive",
        "kolmogorov_time": "positive",
    }
)
def get_stokes_number(
    particle_inertia_time: Union[float, NDArray[np.float64]],
    kolmogorov_time: Union[float, NDArray[np.float64]],
) -> Union[float, NDArray[np.float64]]:
    """
    Calculate the Stokes number.

    The Stokes number (St) is a dimensionless [-] parameter representing the
    relative importance of particle inertia in a fluid flow. A high Stokes
    number indicates that a particle resists following fluid motion, while a
    low Stokes number means the particle closely follows the flow. Given by:

        St = τ_p / τ_k

    - St : Stokes number [-]
    - τ_p (particle_inertia_time) : Particle inertia time [s]
    - τ_k (kolmogorov_time) : Kolmogorov timescale of turbulence [s]

    Arguments:
    ----------
        - particle_inertia_time : Particle inertia time [s]
        - kolmogorov_time : Kolmogorov timescale [s]

    Returns:
    --------
        - Stokes number [-]

    References:
    -----------
        - Wikipedia https://en.wikipedia.org/wiki/Stokes_number
    """
    return particle_inertia_time / kolmogorov_time



"""test for phase_separation.py"""
import numpy as np
from particula.activity.phase_separation import (
    organic_water_single_phase,
    find_phase_sep_index,
    find_phase_separation,
    q_alpha
)


def test_organic_water_single_phase():
    """Test for organic_water_single_phase function."""
    # Test with positive values
    assert np.all(
        organic_water_single_phase([1, 2, 3]) >= 0
    )

    # Test with large values
    assert np.all(
        organic_water_single_phase([1000, 2000, 3000]) >= 0
    )


def test_find_phase_sep_index():
    """Test for find_phase_sep_index function."""
    # Test with positive values
    activity = [.2, .3, .4, .35, .5, .6, .9]
    sep_dic = find_phase_sep_index(activity_data=activity)

    assert sep_dic == {
        'phase_sep_activity': 0,
        'phase_sep_curve': 1,
        'index_phase_sep_starts': 1,
        'index_phase_sep_end': 2
    }

    # Test with values above 1
    activity = [.2, .3, .4, 1, 1.2, 1.3, 1.2, 1.1, 1.0]
    sep_dic = find_phase_sep_index(activity_data=activity)
    assert sep_dic == {
        'phase_sep_activity': 1,
        'phase_sep_curve': 1,
        'index_phase_sep_starts': 4,
        'index_phase_sep_end': 8
    }


def test_find_phase_separation():
    """Test for find_phase_separation function."""
    # Test with positive values

    activity_water = np.array([.2, .3, .4, .35, .5, .6, .9])
    activity_org = np.array([.8, .7, .6, .65, .5, .4, .1])
    sep_dic = find_phase_separation(
        activity_water=activity_water,
        activity_org=activity_org
    )

    assert sep_dic == {
        'phase_sep_check': 1,
        'lower_seperation_index': 1,
        'upper_seperation_index': 2,
        'matching_upper_seperation_index': 4,
        'lower_seperation': 0.3,
        'upper_seperation': 0.4,
        'matching_upper_seperation': 0.5
    }


def test_q_alpha():
    """Test for q_alpha function."""
    # Test with positive values
    assert np.all(
        q_alpha(
            seperation_activity=0.5,
            activities=np.array([.2, .3, .4, .35, .5, .6, .9])) >= 0
    )



"""
Calculates the wall loss rate of a particle in a chamber with a given geometry.
"""

from typing import Union, Tuple
from numpy.typing import NDArray
import numpy as np

from particula.dynamics.properties.wall_loss_coefficient import (
    rectangle_wall_loss_coefficient_via_system_state,
    spherical_wall_loss_coefficient_via_system_state,
)


# pylint: disable=too-many-positional-arguments, too-many-arguments
def spherical_wall_loss_rate(
    wall_eddy_diffusivity: float,
    particle_radius: Union[float, NDArray[np.float64]],
    particle_density: Union[float, NDArray[np.float64]],
    particle_concentration: Union[float, NDArray[np.float64]],
    temperature: float,
    pressure: float,
    chamber_radius: float,
) -> Union[float, NDArray[np.float64]]:
    """
    Calculate the wall loss rate of particles in a spherical chamber.

    This function computes the rate at which particles are lost to the walls
    of a spherical chamber, based on the system state. It uses the wall eddy
    diffusivity, particle properties (radius, density, concentration), and
    environmental conditions (temperature, pressure) to determine the loss
    rate.

    Arguments:
        wall_eddy_diffusivity: The rate of wall eddy diffusivity in inverse
            seconds (s⁻¹).
        particle_radius: The radius of the particle in meters (m).
        particle_density: The density of the particle in kilograms per cubic
            meter (kg/m³).
        particle_concentration: The concentration of particles in the chamber
            in particles per cubic meter (#/m³).
        temperature: The temperature of the system in Kelvin (K).
        pressure: The pressure of the system in Pascals (Pa).
        chamber_radius: The radius of the spherical chamber in meters (m).

    Returns:
        The wall loss rate of the particles in the chamber.
    """

    # Step 1: Calculate the wall loss coefficient
    loss_coefficient = spherical_wall_loss_coefficient_via_system_state(
        wall_eddy_diffusivity=wall_eddy_diffusivity,
        particle_radius=particle_radius,
        particle_density=particle_density,
        temperature=temperature,
        pressure=pressure,
        chamber_radius=chamber_radius,
    )

    # Step 2: Calculate and return the wall loss rate
    return -loss_coefficient * particle_concentration


# pylint: disable=too-many-positional-arguments, too-many-arguments
def rectangle_wall_loss_rate(
    wall_eddy_diffusivity: float,
    particle_radius: Union[float, NDArray[np.float64]],
    particle_density: Union[float, NDArray[np.float64]],
    particle_concentration: Union[float, NDArray[np.float64]],
    temperature: float,
    pressure: float,
    chamber_dimensions: Tuple[float, float, float],
) -> Union[float, NDArray[np.float64]]:
    """
    Calculate the wall loss rate of particles in a rectangular chamber.

    This function computes the rate at which particles are lost to the walls
    of a rectangular chamber, based on the system state. It uses the wall eddy
    diffusivity, particle properties (radius, density, concentration), and
    environmental conditions (temperature, pressure) to determine the loss
    rate. The chamber dimensions (length, width, height) are also taken
    into account.

    Arguments:
        wall_eddy_diffusivity: The rate of wall eddy diffusivity in inverse
            seconds (s⁻¹).
        particle_radius: The radius of the particle in meters (m).
        particle_density: The density of the particle in kilograms per cubic
            meter (kg/m³).
        particle_concentration: The concentration of particles in the chamber
            in particles per cubic meter (#/m³).
        temperature: The temperature of the system in Kelvin (K).
        pressure: The pressure of the system in Pascals (Pa).
        chamber_dimensions: A tuple containing the length, width, and height
            of the rectangular chamber in meters (m).

    Returns:
        The wall loss rate of the particles in the chamber.
    """

    # Step 1: Calculate the wall loss coefficient
    loss_coefficient = rectangle_wall_loss_coefficient_via_system_state(
        wall_eddy_diffusivity=wall_eddy_diffusivity,
        particle_radius=particle_radius,
        particle_density=particle_density,
        temperature=temperature,
        pressure=pressure,
        chamber_dimensions=chamber_dimensions,
    )

    # Step 2: Calculate and return the wall loss rate
    return -loss_coefficient * particle_concentration



"""Test the species_density module if it runs without errors."""

import numpy as np
from particula.activity import species_density


def test_organic_density_estimate():
    """Test the organic_density_estimate function."""
    # Test case 1: Known values
    molar_mass = 180.16
    oxygen2carbon = 0.5
    hydrogen2carbon = 1.0
    nitrogen2carbon = 0.1

    assert species_density.organic_density_estimate(
        molar_mass,
        oxygen2carbon,
        hydrogen2carbon,
        nitrogen2carbon) > 0.0

    # Test case 2: Unknown hydrogen2carbon
    molar_mass = 200.0
    oxygen2carbon = 0.4
    nitrogen2carbon = 0.2

    assert species_density.organic_density_estimate(
        molar_mass,
        oxygen2carbon,
        nitrogen2carbon) > 0.0

    # Test case 3: No nitrogen2carbon
    molar_mass = 150.0
    oxygen2carbon = 0.6
    hydrogen2carbon = 0.8

    assert species_density.organic_density_estimate(
        molar_mass,
        oxygen2carbon,
        hydrogen2carbon) > 0.0

    # Test case 4: Mass ratio conversion
    molar_mass = 18/250.0
    oxygen2carbon = 0.3
    hydrogen2carbon = 1.2
    nitrogen2carbon = 0.05

    assert species_density.organic_density_estimate(
        molar_mass,
        oxygen2carbon,
        hydrogen2carbon,
        nitrogen2carbon,
        mass_ratio_convert=True) > 0.0


def test_organic_array():
    """test the organic_array function."""
    molar_mass_array = [180.16, 200.0, 150.0]
    oxygen2carbon_array = [0.5, 0.4, 0.6]
    hydrogen2carbon_array = [1.0, 1.0, 0.8]
    nitrogen2carbon_array = None

    assert np.all(species_density.organic_array(
        molar_mass=molar_mass_array,
        oxygen2carbon=oxygen2carbon_array,
        hydrogen2carbon=hydrogen2carbon_array,
        nitrogen2carbon=nitrogen2carbon_array) > 0.0)



"""Testing Gas class"""

# pylint: disable=R0801

import numpy as np
from particula.gas.atmosphere import Atmosphere
from particula.gas.species import GasSpecies
from particula.gas.vapor_pressure_strategies import (
    ConstantVaporPressureStrategy,
)


def test_gas_initialization():
    """Test the initialization of a Gas object."""
    vapor_pressure_strategy = ConstantVaporPressureStrategy(
        vapor_pressure=np.array([101325, 101325, 101325])
    )
    names = np.array(["Oxygen1", "Hydrogen1", "Nitrogen1"])
    molar_masses = np.array([0.032, 0.002, 0.028])
    condensables = np.array([False, False, False])
    concentrations = np.array([0.21, 0.79, 0.0])
    gas_species_atmo_test = GasSpecies(
        name=names,
        molar_mass=molar_masses,
        vapor_pressure_strategy=vapor_pressure_strategy,
        condensable=condensables,
        concentration=concentrations,
    )
    # create a gas object
    temperature = 298.15  # Kelvin
    total_pressure = 101325  # Pascals
    gas = Atmosphere(temperature, total_pressure, [gas_species_atmo_test])

    assert gas.temperature == temperature
    assert gas.total_pressure == total_pressure

    # test add_species
    assert len(gas.species) == 1

    # Add a species
    gas.add_species(gas_species=gas_species_atmo_test)
    assert len(gas.species) == 2

    # Remove a species
    gas.remove_species(1)
    assert len(gas.species) == 1



"""
Activity coefficients for organic-water mixtures.

Gorkowski, K., Preston, T. C., &#38; Zuend, A. (2019).
Relative-humidity-dependent organic aerosol thermodynamics
Via an efficient reduced-complexity model.
Atmospheric Chemistry and Physics
https://doi.org/10.5194/acp-19-13383-2019
"""

from typing import Optional, Union, Tuple, List
import numpy as np
from numpy.typing import NDArray
from particula.util.machine_limit import safe_exp
from particula.util.validate_inputs import validate_inputs

from particula.activity.gibbs_mixing import gibbs_mix_weight
from particula.activity.convert_functional_group import (
    convert_to_oh_equivalent
)


@validate_inputs(
    {
        "molar_mass_ratio": "positive",
        "organic_mole_fraction": "nonnegative",
        "oxygen2carbon": "nonnegative",
        "density": "positive",
    }
)
def bat_activity_coefficients(
    molar_mass_ratio: Union[float, NDArray[np.float64]],
    organic_mole_fraction: Union[float, NDArray[np.float64]],
    oxygen2carbon: Union[float, NDArray[np.float64]],
    density: Union[float, NDArray[np.float64]],
    functional_group: Optional[Union[str, List[str]]] = None,
) -> Tuple[
    Union[float, NDArray[np.float64]],
    Union[float, NDArray[np.float64]],
    Union[float, NDArray[np.float64]],
    Union[float, NDArray[np.float64]],
    Union[float, NDArray[np.float64]],
    Union[float, NDArray[np.float64]],
]:
    """
    Calculate the activity coefficients for water and organic matter in
    organic-water mixtures.

    Args:
        - molar_mass_ratio : Ratio of the molecular weight of water to the
          molecular weight of organic matter.
        - organic_mole_fraction : Molar fraction of organic matter in the
          mixture.
        - oxygen2carbon : Oxygen to carbon ratio in the organic compound.
        - density : Density of the mixture, in kg/m^3.
        - functional_group : Optional functional group(s) of the organic
          compound, if applicable.

    Returns:
        - A tuple containing the activity of water, activity
          of organic matter, mass fraction of water, and mass
          fraction of organic matter, gamma_water (activity coefficient),
          and gamma_organic (activity coefficient).
    """
    oxygen2carbon, molar_mass_ratio = convert_to_oh_equivalent(
        oxygen2carbon=oxygen2carbon,
        molar_mass_ratio=molar_mass_ratio,
        functional_group=functional_group,
    )
    gibbs_mix, derivative_gibbs = gibbs_mix_weight(
        molar_mass_ratio=molar_mass_ratio,
        organic_mole_fraction=organic_mole_fraction,
        oxygen2carbon=oxygen2carbon,
        density=density,
    )
    ln_gamma_water = gibbs_mix - organic_mole_fraction * derivative_gibbs
    ln_gamma_org = gibbs_mix + (1.0 - organic_mole_fraction) * derivative_gibbs

    gamma_water = safe_exp(ln_gamma_water)
    gamma_organic = safe_exp(ln_gamma_org)

    activity_water = gamma_water * (1.0 - organic_mole_fraction)
    activity_organic = gamma_organic * organic_mole_fraction

    mass_water = (
        (1.0 - organic_mole_fraction)
        * molar_mass_ratio
        / ((1.0 - organic_mole_fraction) * (molar_mass_ratio - 1.0) + 1.0)
    )
    mass_organic = 1.0 - mass_water

    return (
        activity_water,
        activity_organic,
        mass_water,
        mass_organic,
        gamma_water,
        gamma_organic,
    )



"""
Fluid RMS fluctuation velocity calculation module.
"""

from typing import Union
from numpy.typing import NDArray
import numpy as np

from particula.util.validate_inputs import validate_inputs
from particula.gas.properties.kolmogorov_module import get_kolmogorov_velocity


@validate_inputs(
    {
        "re_lambda": "positive",
        "kinematic_viscosity": "positive",
        "turbulent_dissipation": "positive",
    }
)
def get_fluid_rms_velocity(
    re_lambda: Union[float, NDArray[np.float64]],
    kinematic_viscosity: Union[float, NDArray[np.float64]],
    turbulent_dissipation: Union[float, NDArray[np.float64]],
) -> Union[float, NDArray[np.float64]]:
    """
    Calculate the fluid RMS fluctuation velocity.

    The fluid root-mean-square (RMS) velocity fluctuation quantifies
    turbulence intensity in a fluid flow. It is calculated as:

        u' = (R_λ^(1/2) v_K) / 15^(1/4)

    - u' : Fluid RMS fluctuation velocity [m/s]
    - R_λ (re_lambda) : Taylor-microscale Reynolds number [-]
    - v_K : Kolmogorov velocity scale, computed as v_K = ( ε)^(1/4) [m/s]
    - v (kinematic_viscosity) : Kinematic viscosity of the fluid [m²/s]
    - ε (turbulent_dissipation) : Turbulent energy dissipation rate [m²/s³]


    Arguments:
    ----------
        - re_lambda : Taylor-microscale Reynolds number [-]
        - kinematic_viscosity : Kinematic viscosity of the fluid [m²/s]
        - turbulent_dissipation : Rate of dissipation of turbulent kinetic
            energy [m²/s³]

    Returns:
    --------
        - Fluid RMS fluctuation velocity [m/s]
    """
    kolmogorov_velocity = get_kolmogorov_velocity(
        kinematic_viscosity, turbulent_dissipation
    )
    return (re_lambda**0.5 * kolmogorov_velocity) / (15**0.25)



"""Tests for representation builders."""

import numpy as np

from particula.particles.representation_builders import (
    ParticleMassRepresentationBuilder,
    ParticleRadiusRepresentationBuilder,
    PresetParticleRadiusBuilder,
    ResolvedParticleMassRepresentationBuilder,
    PresetResolvedParticleMassBuilder,
)
from particula.particles.representation import ParticleRepresentation
from particula.particles.distribution_strategies import (
    RadiiBasedMovingBin,
    ParticleResolvedSpeciatedMass,
)
from particula.particles.surface_strategies import SurfaceStrategyVolume
from particula.particles.activity_strategies import ActivityIdealMass


def test_mass_particle_representation_builder():
    """Test MassParticleRepresentationBuilder Builds."""
    builder = ParticleMassRepresentationBuilder()
    builder.set_distribution_strategy(RadiiBasedMovingBin())
    builder.set_activity_strategy(ActivityIdealMass())
    builder.set_surface_strategy(SurfaceStrategyVolume())
    builder.set_mass(np.array([1.0, 2.0, 3.0]))
    builder.set_density(np.array([1.0, 2.0, 3.0]), "kg/m^3")
    builder.set_concentration(np.array([10, 20, 30]))
    builder.set_charge(1.0)
    particle_representation = builder.build()
    assert isinstance(particle_representation, ParticleRepresentation)


def test_radius_particle_representation_builder():
    """Test RadiusParticleRepresentationBuilder Builds."""
    builder = ParticleRadiusRepresentationBuilder()
    builder.set_distribution_strategy(RadiiBasedMovingBin())
    builder.set_activity_strategy(ActivityIdealMass())
    builder.set_surface_strategy(SurfaceStrategyVolume())
    builder.set_radius(np.array([1.0, 2.0, 3.0]))
    builder.set_density(np.array([1.0, 2.0, 3.0]), "kg/m^3")
    builder.set_concentration(np.array([10, 20, 30]))
    builder.set_charge(np.array([1.0, 2.0, 3.0]))
    particle_representation = builder.build()
    assert isinstance(particle_representation, ParticleRepresentation)


def test_limited_radius_particle_builder():
    """Test LimitedRadiusParticleBuilder Builds."""
    # default values
    builder = PresetParticleRadiusBuilder()
    particle_representation_defaults = builder.build()
    assert isinstance(particle_representation_defaults, ParticleRepresentation)

    # set values
    builder = PresetParticleRadiusBuilder()
    builder.set_mode(np.array([100, 2000])*1e-9, "m")
    builder.set_geometric_standard_deviation(np.array([1.4, 1.5]))
    builder.set_number_concentration(np.array([1e3, 1e3])*1e6, "1/m^3")
    particle_representation = builder.build()
    assert isinstance(particle_representation, ParticleRepresentation)


def test_resolved_mass_particle_representation_builder():
    """Test ResolvedMassParticleRepresentationBuilder Builds."""
    builder = ResolvedParticleMassRepresentationBuilder()
    builder.set_distribution_strategy(ParticleResolvedSpeciatedMass())
    builder.set_activity_strategy(ActivityIdealMass())
    builder.set_surface_strategy(SurfaceStrategyVolume())
    builder.set_mass(np.array([1.0, 2.0, 3.0]))
    builder.set_density(np.array([1.0, 2.0, 3.0]), "kg/m^3")
    builder.set_charge(1.0)
    builder.set_volume(1)
    particle_representation = builder.build()
    assert isinstance(particle_representation, ParticleRepresentation)


def test_preset_resolved_mass_particle_builder():
    """Test PresetResolvedMassParticleBuilder Builds."""
    # default values
    builder = PresetResolvedParticleMassBuilder()
    particle_representation_defaults = builder.build()
    assert isinstance(particle_representation_defaults, ParticleRepresentation)

    # set values
    builder = PresetResolvedParticleMassBuilder()
    builder.set_mode(np.array([100, 2000])*1e-9, "m")
    builder.set_geometric_standard_deviation(np.array([1.4, 1.5]))
    builder.set_particle_resolved_count(1000)
    builder.set_volume(1.0)
    particle_representation = builder.build()
    assert isinstance(particle_representation, ParticleRepresentation)



"""Test Brownian module functions.
"""

import pytest
import numpy as np
from particula.dynamics.coagulation import brownian_kernel
from particula.util.constants import BOLTZMANN_CONSTANT


def test_mean_free_path_l_single_value():
    """
    Test mean_free_path_l with single float inputs for diffusivity and mean
    thermal speed.
    """
    diffusivity_particle = 1.0e-5  # example diffusivity [m^2/s]
    mean_thermal_speed_particle = 100  # example speed [m/s]
    expected_path = (
        8 * diffusivity_particle / (np.pi * mean_thermal_speed_particle)
    )
    value = brownian_kernel.mean_free_path_l(
        diffusivity_particle, mean_thermal_speed_particle
    )
    assert np.isclose(value, expected_path)


def test_mean_free_path_l_array_input():
    """
    Test mean_free_path_l with numpy array inputs for diffusivity and mean
    thermal speed.
    """
    diffusivity_particle = np.array([1.0e-5, 1.0e-4])
    mean_thermal_speed_particle = np.array([100, 200])
    expected_path = (
        8 * diffusivity_particle / (np.pi * mean_thermal_speed_particle)
    )
    value = brownian_kernel.mean_free_path_l(
        diffusivity_particle, mean_thermal_speed_particle
    )
    np.testing.assert_allclose(value, expected_path, rtol=1e-6)


def test_mean_free_path_l_input_validation():
    """
    Ensure that providing incorrect input types to mean_free_path_l raises a
    TypeError.
    """
    with pytest.raises(TypeError):
        brownian_kernel.mean_free_path_l("not a number", "also not a number")


def test_g_collection_term_single_value():
    """
    Test g_collection_term with single float inputs for mean free path and
    radius of particles.
    """
    mean_free_path_particle = 6.085302617379366e-08  # mean free path in meters
    radius_particle = 1e-5  # radius in meters
    expected_value = (
        (2 * radius_particle + mean_free_path_particle) ** 3
        - (4 * radius_particle**2 + mean_free_path_particle**2) ** (3 / 2)
    ) / (6 * radius_particle * mean_free_path_particle) - 2 * radius_particle
    value = brownian_kernel.g_collection_term(
        mean_free_path_particle, radius_particle
    )
    assert np.isclose(value, expected_value)


def test_g_collection_term_array_input():
    """
    Test g_collection_term with numpy array inputs for mean free path and
    radius of particles.
    """
    mean_free_path_particle = np.array([0.0005, 0.0005])
    radius_particle = np.array([0.0001, 0.0002])
    expected_value = (
        (2 * radius_particle + mean_free_path_particle) ** 3
        - (4 * radius_particle**2 + mean_free_path_particle**2) ** (3 / 2)
    ) / (6 * radius_particle * mean_free_path_particle) - 2 * radius_particle
    value = brownian_kernel.g_collection_term(
        mean_free_path_particle, radius_particle
    )
    np.testing.assert_allclose(value, expected_value, rtol=1e-6)


def test_g_collection_term_input_validation():
    """
    Ensure that providing incorrect input types to g_collection_term raises
    a TypeError.
    """
    with pytest.raises(TypeError):
        brownian_kernel.g_collection_term("not a number", "also not a number")


def test_g_collection_term_zero_radius():
    """
    Test that g_collection_term handles a case where the radius is zero
    without crashing.
    """
    mean_free_path_particle = 0.0005
    radius_particle = 0.0
    with pytest.raises(ZeroDivisionError):
        brownian_kernel.g_collection_term(
            mean_free_path_particle, radius_particle
        )


def test_brownian_diffusivity_single_value():
    """
    Test brownian_diffusivity with single float inputs for temperature and
    aerodynamic mobility.
    """
    temperature = 298  # example temperature in Kelvin
    aerodynamic_mobility = 1.5e-9  # example mobility in m^2/s
    expected_diffusivity = (
        float(BOLTZMANN_CONSTANT) * temperature * aerodynamic_mobility
    )
    value = brownian_kernel.brownian_diffusivity(
        temperature, aerodynamic_mobility
    )
    assert np.isclose(value, expected_diffusivity)


def test_brownian_diffusivity_array_input():
    """
    Test brownian_diffusivity with numpy array inputs for temperature and
    aerodynamic mobility.
    """
    temperature = np.array([298, 300])
    aerodynamic_mobility = np.array([1.5e-9, 2.0e-9])
    expected_diffusivity = (
        float(BOLTZMANN_CONSTANT) * temperature * aerodynamic_mobility
    )
    value = brownian_kernel.brownian_diffusivity(
        temperature, aerodynamic_mobility
    )
    np.testing.assert_allclose(value, expected_diffusivity, rtol=1e-6)


def test_brownian_coagulation_kernel_basic():
    """
    Test brownian_coagulation_kernel with basic input values.
    """
    radius_particle = np.array([1e-9, 2e-9])  # radii in meters
    diffusivity_particle = np.array([1e-12, 1e-12])  # diffusivity in m^2/s
    g_collection_term_particle = np.array([0.5, 0.5])  # dimensionless
    mean_thermal_speed_particle = np.array([100, 100])  # speed in m/s
    alpha_collision_efficiency = np.array([1.0, 1.0])  # dimensionless

    # Calculate expected values
    expected_kernel = np.array(
        [[1.77697548e-15, 3.99769516e-15], [3.99769516e-15, 7.10577039e-15]]
    )

    result_kernel = brownian_kernel.brownian_coagulation_kernel(
        radius_particle,
        diffusivity_particle,
        g_collection_term_particle,
        mean_thermal_speed_particle,
        alpha_collision_efficiency,
    )

    np.testing.assert_allclose(result_kernel, expected_kernel, rtol=1e-6)


def test_brownian_coagulation_kernel_with_defaults():
    """
    Test that default parameters are handled correctly.
    """
    radius_particle = np.array([1e-9])  # single radius
    diffusivity_particle = np.array([1e-12])  # single diffusivity
    g_collection_term_particle = np.array([0.5])  # single collection term
    mean_thermal_speed_particle = np.array([100])  # single speed

    # expected result for a single input with default alpha
    expected_kernel = np.array([1.77697548e-15])

    result_kernel = brownian_kernel.brownian_coagulation_kernel(
        radius_particle,
        diffusivity_particle,
        g_collection_term_particle,
        mean_thermal_speed_particle,
    )
    assert np.isclose(result_kernel, expected_kernel)


def test_brownian_coagulation_kernel_input_validation():
    """
    Ensure that providing incorrect input types raises a TypeError.
    """
    with pytest.raises(TypeError):
        brownian_kernel.brownian_coagulation_kernel(
            "not a number", "not a number", "not a number", "not a number"
        )


def test_brownian_coagulation_kernel_via_system_state_basic():
    """
    Test the complete system function with basic input values.
    """
    # diameters, 2 nm, 1 um, 20 um
    radius_particle = np.array([1e-9, 5e-7, 10e-6])  # radii in meters
    mass_particle = 4 / 3 * np.pi * radius_particle**3 * 1000  # mass in kg
    temperature = 298  # temperature in Kelvin
    pressure = 101325  # pressure in Pascal
    alpha_collision_efficiency = np.array([1.0, 1.0, 1.0])  # dimensionless

    # Seinfeld and Pandis Table 13.3, larger particles seem off by 2x
    # table 13.3
    # [8.9e-16, 7.8e-12, 17e-9],
    # [7.8e-12, 6.9e-16, 3.9e-15],
    # [17e-9, 3.9e-15, 6.0e-16]
    expected_output = np.array(
        [
            [8.88541843e-16, 7.57666671e-12, 1.64739803e-10],
            [7.57666671e-12, 6.80940034e-16, 3.82064762e-15],
            [1.64739803e-10, 3.82064762e-15, 5.99655468e-16],
        ]
    )

    # Call the system function
    result = brownian_kernel.brownian_coagulation_kernel_via_system_state(
        radius_particle,
        mass_particle,
        temperature,
        pressure,
        alpha_collision_efficiency,
    )

    # Assert to check if the calculated results match expected results closely
    np.testing.assert_allclose(result, expected_output, rtol=1e-6)


def test_brownian_coagulation_kernel_via_system_state_input_validation():
    """
    Ensure that providing incorrect input types or sizes raises appropriate
    errors.
    """
    with pytest.raises(TypeError):
        brownian_kernel.brownian_coagulation_kernel_via_system_state(
            radius_particle="not a number",
            mass_particle="not a number",
            temperature="not a number",
            pressure="not a number",
            alpha_collision_efficiency="not a number",
        )



""" Module for calculate slip correction
"""

from typing import Union
from numpy.typing import NDArray
import numpy as np


def cunningham_slip_correction(
    knudsen_number: Union[float, NDArray[np.float64]]
) -> Union[float, NDArray[np.float64]]:
    """
    Calculate the Cunningham slip correction factor. Accounts for
    non-continuum effects on small particles.

    Args:
    -----
    - knudsen_number: Knudsen number [unitless].

    Returns:
    --------
    - Slip correction factor [unitless].

    Reference:
    ----------
    - Dimensionless quantity accounting for non-continuum effects
    on small particles. It is a deviation from Stokes' Law.
    Stokes assumes a no-slip condition that is not correct at
    high Knudsen numbers. The slip correction factor is used to
    calculate the friction factor.
    Thus, the slip correction factor is about unity (1) for larger
    particles (Kn -> 0). Its behavior on the other end of the
    spectrum (smaller particles; Kn -> inf) is more nuanced, though
    it tends to scale linearly on a log-log scale, log Cc vs log Kn.
    - https://en.wikipedia.org/wiki/Cunningham_correction_factor
    """
    return 1 + knudsen_number * (1.257 + 0.4 * np.exp(-1.1 / knudsen_number))
